# coding=utf-8
# Filename: test_core.py
# pylint: disable=C0111,E1003,R0904,C0103,R0201,C0102
from __future__ import division, absolute_import, print_function

from km3pipe.testing import TestCase, StringIO, MagicMock, patch
from km3pipe.core import Pipeline, Module, Pump, Blob, Geometry
from km3pipe.dataclasses import HitSeries

import numpy as np

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestPipeline(TestCase):
    """Tests for the main pipeline"""

    def setUp(self):
        self.pl = Pipeline()

    def test_attach(self):
        self.pl.attach(Module, 'module1')
        self.pl.attach(Module, 'module2')
        print([m.name for m in self.pl.modules])
        self.assertEqual('module1', self.pl.modules[0].name)
        self.assertEqual('module2', self.pl.modules[1].name)

    def test_attach_bundle(self):
        modules = [Module, Module]
        self.pl.attach_bundle(modules)
        self.assertEqual(2, len(self.pl.modules))

    def test_attach_function(self):
        self.pl.attach(lambda x: 1)
        self.pl.attach(lambda x: 2, "Another Lambda")
        self.assertEqual('<lambda>', self.pl.modules[0].name)
        self.assertEqual('Another Lambda', self.pl.modules[1].name)

    def test_drain_calls_each_attached_module(self):
        pl = Pipeline(blob=1)

        func_module_spy = MagicMock()

        def func_module(blob):
            func_module_spy()
            return blob

        pl.attach(Module, 'module1')
        pl.attach(func_module, 'module2')
        pl.attach(Module, 'module3')

        for module in pl.modules:
            print(module)
            if isinstance(module, Module):
                print("normal module, mocking")
                module.process = MagicMock(return_value={})

        n = 3

        pl.drain(n)

        for module in pl.modules:
            try:
                # Regular modules
                self.assertEqual(n, module.process.call_count)
            except AttributeError:
                # Function module
                self.assertEqual(n, func_module_spy.call_count)

    def test_drain_calls_process_method_on_each_attached_module(self):
        pl = Pipeline(blob=1)

        pl.attach(Module, 'module1')
        pl.attach(Module, 'module2')
        pl.attach(Module, 'module3')
        for module in pl.modules:
            module.process = MagicMock(return_value={})
        n = 3
        pl.drain(n)
        for module in pl.modules:
            self.assertEqual(n, module.process.call_count)

    def test_drain_doesnt_call_process_if_blob_is_none(self):
        pl = Pipeline(blob=1)

        pl.attach(Module, 'module1')
        pl.attach(Module, 'module2')
        pl.attach(Module, 'module3')
        pl.modules[0].process = MagicMock(return_value=None)
        pl.modules[1].process = MagicMock(return_value={})
        pl.modules[2].process = MagicMock(return_value={})
        n = 3
        pl.drain(n)
        self.assertEqual(n, pl.modules[0].process.call_count)
        self.assertEqual(0, pl.modules[1].process.call_count)
        self.assertEqual(0, pl.modules[2].process.call_count)

    def test_conditional_module_not_called_if_key_not_in_blob(self):
        pl = Pipeline(blob=1)

        pl.attach(Module, 'module1')
        pl.attach(Module, 'module2', only_if='foo')
        pl.attach(Module, 'module3')

        for module in pl.modules:
            module.process = MagicMock(return_value={})

        pl.drain(1)

        self.assertEqual(1, pl.modules[0].process.call_count)
        self.assertEqual(0, pl.modules[1].process.call_count)
        self.assertEqual(1, pl.modules[2].process.call_count)

    def test_conditional_module_called_if_key_in_blob(self):
        pl = Pipeline(blob=1)

        pl.attach(Module, 'module1')
        pl.attach(Module, 'module2', only_if='foo')
        pl.attach(Module, 'module3')

        pl.modules[0].process = MagicMock(return_value={'foo': 23})
        pl.modules[1].process = MagicMock(return_value={})
        pl.modules[2].process = MagicMock(return_value={})

        pl.drain(1)

        self.assertEqual(1, pl.modules[0].process.call_count)
        self.assertEqual(1, pl.modules[1].process.call_count)
        self.assertEqual(1, pl.modules[2].process.call_count)

    def test_drain_calls_function_modules(self):
        pl = Pipeline(blob=1)

        func_module1 = MagicMock()
        func_module2 = MagicMock()
        func_module3 = MagicMock()

        func_module1.__name__ = "MagicMock"
        func_module2.__name__ = "MagicMock"
        func_module3.__name__ = "MagicMock"

        pl.attach(func_module1, 'module1')
        pl.attach(func_module2, 'module2')
        pl.attach(func_module3, 'module3')
        pl.drain(1)
        self.assertEqual(1, pl.modules[0].call_count)
        self.assertEqual(1, pl.modules[1].call_count)
        self.assertEqual(1, pl.modules[2].call_count)

    def test_finish(self):
        self.pl.finish()

    def test_drain_calls_finish_on_each_attached_module(self):
        self.pl.attach(Module, 'module1')
        self.pl.attach(Module, 'module2')
        self.pl.attach(lambda x: 1, 'func_module')
        for module in self.pl.modules:
            module.finish = MagicMock()
        self.pl.drain(4)
        for module in self.pl.modules:
            if module.name != 'func_module':
                self.assertEqual(1, module.finish.call_count)

    def test_ctrl_c_handling(self):
        pl = Pipeline()
        self.assertFalse(pl._stop)
        pl._handle_ctrl_c()  # first KeyboardInterrupt
        self.assertTrue(pl._stop)
        with self.assertRaises(SystemExit):
            pl._handle_ctrl_c()  # second KeyboardInterrupt


class TestModule(TestCase):
    """Tests for the pipeline module"""

    def test_name_can_be_set_on_init(self):
        name = 'foo'
        module = Module(name=name)
        self.assertEqual(name, module.name)

    def test_name_is_read_only(self):
        module = Module(name='foo')
        with self.assertRaises(AttributeError):
            module.name = 'narf'

    def test_process(self):
        blob = Blob()
        module = Module(name='foo')
        processed_blob = module.process(blob)
        self.assertIs(blob, processed_blob)

    def test_add_parameter(self):
        module = Module()
        module.add('foo', 'default')
        self.assertDictEqual({'foo': 'default'}, module.parameters)

    def test_get_parameter(self):
        module = Module()
        module.add('foo', 'default')
        self.assertEqual('default', module.get('foo'))

    def test_default_parameter_value_can_be_overwritten(self):
        class Foo(Module):
            def __init__(self, **context):
                super(self.__class__, self).__init__(**context)
                self.foo = self.get('foo') or 'default_foo'
        module = Foo()
        self.assertEqual('default_foo', module.foo)
        module = Foo(foo='overwritten')
        self.assertEqual('overwritten', module.foo)

    def test_finish(self):
        module = Module()
        module.finish()


class TestPump(TestCase):
    """Tests for the pump"""

    def test_rewind_file(self):
        pump = Pump()
        test_file = StringIO("Some content")
        pump.blob_file = test_file
        pump.blob_file.read(1)
        self.assertEqual(1, pump.blob_file.tell())
        pump.rewind_file()
        self.assertEqual(0, pump.blob_file.tell())


class TestBlob(TestCase):
    """Tests for the blob holding the data"""

    def test_field_can_be_added(self):
        blob = Blob()
        blob['foo'] = 1
        self.assertEqual(1, blob['foo'])


class TestGeometry(TestCase):
    """Tests for the Geometry class"""

    def test_init_with_wrong_file_extension(self):
        with self.assertRaises(NotImplementedError):
            geo = Geometry(filename='foo')

    @patch('km3pipe.core.Detector')
    def test_init_with_filename(self, mock_detector):
        geo = Geometry(filename='foo.detx')
        mock_detector.assert_called_with(filename='foo.detx')

    @patch('km3pipe.core.Detector')
    def test_init_with_det_id(self, mock_detector):
        geo = Geometry(det_id=1)
        mock_detector.assert_called_with(t0set=None, calibration=None, det_id=1)
        geo = Geometry(det_id=1, calibration=2, t0set=3)
        mock_detector.assert_called_with(t0set=3, calibration=2, det_id=1)

    def test_apply_to_list(self):
        geo = Geometry()
        hits = [1, 2, 3]
        geo._apply_to_hitseries = MagicMock()
        geo.apply(hits)
        geo._apply_to_hitseries.assert_called_with(hits)

    def test_apply_to_hitseries(self):

        class FakeDetector(object):
            def pmt_with_id(self, i):
                pmt = MagicMock(dir=np.array((i*10+i, i*10+i+1, i*10+i+2)),
                                pos=np.array((i*100+i, i*100+i+1, i*100+i+2)),
                                t0=1000*i)
                return pmt

        geo = Geometry(detector=FakeDetector())

        n = 5
        ids = np.arange(n)
        dom_ids = np.arange(n)
        dir_xs = np.arange(n)
        dir_ys = np.arange(n)
        dir_zs = np.arange(n)
        times = np.arange(n)
        tots = np.arange(n)
        channel_ids = np.arange(n)
        triggereds = np.ones(n)
        pmt_ids = np.arange(n)
        t0s = np.arange(n)
        pos_xs = np.arange(n)
        pos_ys = np.arange(n)
        pos_zs = np.arange(n)

        hits = HitSeries.from_arrays(
            channel_ids,
            dir_xs,
            dir_ys,
            dir_zs,
            dom_ids,
            ids,
            pmt_ids,
            pos_xs,
            pos_ys,
            pos_zs,
            t0s,
            times,
            tots,
            triggereds,
            0,      # event_id
        )

        self.assertEqual(0, hits[0].time)
        self.assertEqual(4, hits[4].time)
        self.assertFalse(np.isnan(hits[2].pos_y))

        geo._apply_to_hitseries(hits)

        self.assertAlmostEqual(303, hits[3].pos_x)
        self.assertAlmostEqual(304, hits[3].pos_y)
        self.assertAlmostEqual(305, hits[3].pos_z)
        self.assertAlmostEqual(406, hits[4].pos_z)
        self.assertAlmostEqual(2, hits[0].dir_z)
        self.assertAlmostEqual(12, hits[1].dir_y)
        self.assertAlmostEqual(22, hits[2].dir_x)

        self.assertEqual(1001, hits[1].time)
        self.assertEqual(4004, hits[4].time)

        for idx, hit in enumerate(hits):
            h = hit
            if idx == 3:
                break

        self.assertAlmostEqual(303.0, h.pos_x)



#!/usr/bin/env python
# coding=utf-8
# Filename: __version__.py
# pylint: disable=C0103
"""
Pep 386 compliant version info.

    (major, minor, micro, alpha/beta/rc/final, #)
    (1, 1, 2, 'alpha', 0) => "1.1.2.dev"
    (1, 2, 0, 'beta', 2) => "1.2b2"

"""
import json
try:
    from urllib import urlopen
except ImportError:
    from urllib.request import urlopen

try:
    from km3pipe.config import Config
    config = Config()
except ImportError:
    config = None

from km3pipe.logger import logging

__author__ = 'tamasgal'

log = logging.getLogger(__name__)  # pylint: disable=C0103

version_info = (6, 4, 3, 'final', 0)


def _get_version(version_info):
    """Return a PEP 386-compliant version number."""
    assert len(version_info) == 5
    assert version_info[3] in ('alpha', 'beta', 'rc', 'final')

    parts = 2 if version_info[2] == 0 else 3
    main = '.'.join(map(str, version_info[:parts]))

    sub = ''
    if version_info[3] == 'alpha' and version_info[4] == 0:
        sub = '.dev'
    elif version_info[3] != 'final':
        mapping = {'alpha': 'a', 'beta': 'b', 'rc': 'c'}
        sub = mapping[version_info[3]] + str(version_info[4])

    return str(main + sub)


def _get_latest_version():
    response = urlopen('https://pypi.python.org/pypi/km3pipe/json')
    content = response.read()
    latest_version = json.loads(content.decode())['info']['version']
    return str(latest_version)


def check_for_update():
    try:
        latest_version = _get_latest_version()
    except IOError:
        pass
    else:
        version = _get_version(version_info)
        if latest_version != version:
            log.warn("There is an update for km3pipe available.\n" +
                     "    Installed: {0}\n"
                     "    Latest: {1}\n".format(version, latest_version) +
                     "Please run `pip install --upgrade km3pipe` to update.")


version = _get_version(version_info)
if config is not None:
    if config.check_for_updates:
        check_for_update()



# coding=utf-8
# Filename: root.py
# pylint: disable=C0103,R0903
# vim:set ts=4 sts=4 sw=4 et:
"""
Read and write Vanilla ROOT files.
"""
from __future__ import division, absolute_import, print_function

from collections import defaultdict
from six import string_types

import numpy as np
import rootpy as rp
import root_numpy as rnp
import rootpy.ROOT as ROOT
from rootpy.io import root_open
from rootpy.plotting import Hist, Hist2D, Hist3D
from rootpy.tree import Tree, TreeModel, FloatCol, IntCol

import km3pipe as kp
from km3pipe.core import Pump, Module
from km3pipe.dataclasses import KM3Array, deserialise_map
from km3pipe.logger import logging
from km3pipe.tools import camelise, decamelise, insert_prefix_to_dtype, split

log = logging.getLogger(__name__)  # pylint: disable=C0103

__author__ = "Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def open_rfile(rfile, fmode='r'):
    if isinstance(rfile, string_types):
        return root_open(rfile, mode=fmode)
    return rfile


def get_hist(rfile, histname, get_overflow=False):
    """Read a 1D Histogram."""
    rfile = open_rfile(rfile)
    hist = rfile[histname]
    xlims = np.array(list(hist.xedges()))
    bin_values = rnp.hist2array(hist, include_overflow=get_overflow)
    #bin_errors = list(hist.xerror())
    rfile.close()
    return bin_values, xlims#, bin_errors


def get_hist2d(rfile, histname, get_overflow=False):
    """Read a 2D Histogram."""
    rfile = open_rfile(rfile)
    hist = rfile[histname]
    xlims = np.array(list(hist.xedges()))
    ylims = np.array(list(hist.yedges()))
    bin_values = rnp.hist2array(hist, include_overflow=get_overflow)
    #bin_errors = list(hist.xerror())
    rfile.close()
    return bin_values, xlims, ylims#, bin_errors


def get_hist3d(rfile, histname, get_overflow=False):
    """Read a 3D Histogram."""
    rfile = open_rfile(rfile)
    hist = rfile[histname]
    xlims = np.array(list(hist.xedges()))
    ylims = np.array(list(hist.xedges()))
    zlims = np.array(list(hist.zedges()))
    bin_values = rnp.hist2array(hist, include_overflow=get_overflow)
    #bin_errors = list(hist.xerror())
    rfile.close()
    return bin_values, xlims, ylims, zlims#, bin_errors



# coding=utf-8
# Filename: daq.py
# pylint: disable=R0903
"""
Pumps for the DAQ data formats.

"""
from __future__ import division, absolute_import, print_function

from collections import defaultdict
from io import BytesIO
import math
import struct
from struct import unpack
import pprint

import numpy as np

from km3pipe.core import Pump, Module, Blob
from km3pipe.dataclasses import EventInfo, HitSeries
from km3pipe.tools import ignored
from km3pipe.logger import logging

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = logging.getLogger(__name__)  # pylint: disable=C0103

DATA_TYPES = {
    101: 'DAQSuperFrame',
    201: 'DAQSummaryFrame',
    1001: 'DAQTimeslice',
    2001: 'DAQSummaryslice',
    10001: 'DAQEvent',
}
MINIMAL_RATE_HZ = 2.0e3
MAXIMAL_RATE_HZ = 2.0e6



class TimesliceParser(Module):
    """Preliminary parser for DAQTimeslice"""
    def process(self, blob):
        if str(blob['CHPrefix'].tag) != 'IO_TSL':
            log.error("Not an IO_TSL blob")
            return

        try:
            data = BytesIO(blob['CHData'])
            tsl_size, datatype = unpack('<ii', data.read(8))
            det_id, run, sqnr = unpack('<iii', data.read(12))
            timestamp, ns_ticks, n_frames = unpack('<iii', data.read(12))

            ts_frames = blob['TimesliceFrames'] = defaultdict(list)

            for i in range(n_frames):
                frame_size, datatype = unpack('<ii', data.read(8))
                det_id, run, sqnr = unpack('<iii', data.read(12))
                timestamp, ns_ticks, dom_id = unpack('<iii', data.read(12))
                dom_status = unpack('<iiiii', data.read(5*4))
                n_hits = unpack('<i', data.read(4))[0]
                hits = []
                for j in range(n_hits):
                    hit = unpack('!BlB', data.read(6))
                    ts_frames[dom_id].append(hit)
        except struct.error:
            log.error("Could not parse Timeslice")
            log.error(blob.keys())
        else:
            return blob


class DAQPump(Pump):
    """A pump for binary DAQ files."""

    def __init__(self, **context):
        super(self.__class__, self).__init__(**context)
        self.filename = self.get('filename')
        self.frame_positions = []
        self.index = 0

        if self.filename:
            self.open_file(self.filename)
            self.determine_frame_positions()
        else:
            log.warn("No filename specified. Take care of the file handling!")

    def next_blob(self):
        """Get the next frame from file"""
        blob_file = self.blob_file
        try:
            preamble = DAQPreamble(file_obj=blob_file)
        except struct.error:
            raise StopIteration

        try:
            data_type = DATA_TYPES[preamble.data_type]
        except KeyError:
            log.error("Unkown datatype: {0}".format(preamble.data_type))
            data_type = 'Unknown'

        blob = Blob()
        blob[data_type] = None
        blob['DAQPreamble'] = preamble

        if data_type == 'DAQSummaryslice':
            daq_frame = DAQSummaryslice(blob_file)
            blob[data_type] = daq_frame
            blob['DAQHeader'] = daq_frame.header
        elif data_type == 'DAQEvent':
            daq_frame = DAQEvent(blob_file)
            blob[data_type] = daq_frame
            blob['DAQHeader'] = daq_frame.header
        else:
            log.warning("Skipping DAQ frame with data type code '{0}'."
                        .format(preamble.data_type))
            blob_file.seek(preamble.length - DAQPreamble.size, 1)

        return blob

    def seek_to_frame(self, index):
        """Move file pointer to the frame with given index."""
        pointer_position = self.frame_positions[index]
        self.blob_file.seek(pointer_position, 0)

    def get_blob(self, index):
        """Return blob at given index."""
        self.seek_to_frame(index)
        return self.next_blob()

    def determine_frame_positions(self):
        """Record the file pointer position of each frame"""
        self.rewind_file()
        with ignored(struct.error):
            while True:
                pointer_position = self.blob_file.tell()
                length = struct.unpack('<i', self.blob_file.read(4))[0]
                self.blob_file.seek(length - 4, 1)
                self.frame_positions.append(pointer_position)
        self.rewind_file()
        log.info("Found {0} frames.".format(len(self.frame_positions)))

    def process(self, blob):
        """Pump the next blob to the modules"""
        return self.next_blob()

    def finish(self):
        """Clean everything up"""
        self.blob_file.close()

    def __len__(self):
        if not self.frame_positions:
            self.determine_frame_positions()
        return len(self.frame_positions)

    def __iter__(self):
        return self

    def next(self):
        """Python 2/3 compatibility for iterators"""
        return self.__next__()

    def __next__(self):
        try:
            blob = self.get_blob(self.index)
        except IndexError:
            self.index = 0
            raise StopIteration
        self.index += 1
        return blob

    def __getitem__(self, index):
        if isinstance(index, int):
            return self.get_blob(index)
        elif isinstance(index, slice):
            return self._slice_generator(index)
        else:
            raise TypeError("index must be int or slice")

    def _slice_generator(self, index):
        """A simple slice generator for iterations"""
        start, stop, step = index.indices(len(self))
        for i in range(start, stop, step):
            yield self.get_blob(i)


class DAQProcessor(Module):
    def __init__(self, **context):
        super(self.__class__, self).__init__(**context)
        self.index = 0

    def process(self, blob):
        tag = str(blob['CHPrefix'].tag)
        data = blob['CHData']

        if tag == 'IO_EVT':
            self.process_event(data, blob)
        if tag == 'IO_SUM':
            self.process_summaryslice(data, blob)

        return blob

    def process_event(self, data, blob):
        data_io = BytesIO(data)
        preamble = DAQPreamble(file_obj=data_io)  # noqa
        event = DAQEvent(file_obj=data_io)
        header = event.header

        hits = event.snapshot_hits
        n_hits = event.n_snapshot_hits
        dom_ids, channel_ids, times, tots = zip(*hits)
        zeros = np.zeros(n_hits)
        nans = np.full_like(zeros, np.nan)
        triggereds = np.zeros(n_hits)
        triggered_map = {}
        for triggered_hit in event.triggered_hits:
            dom_id, pmt_id, time, tot, _ = triggered_hit
            triggered_map[(dom_id, pmt_id, time, tot)] = True
        for idx, hit in enumerate(hits):
            triggereds[idx] = hit in triggered_map

        hit_series = HitSeries.from_arrays(
            channel_ids,
            nans,  # dir_x
            nans,  # dir_y
            nans,  # dir_z
            dom_ids,
            range(n_hits),  # id
            zeros,  # pmt_id
            nans,  # pos_x
            nans,  # pos_y
            nans,  # pos_z
            zeros,  # t0
            times,
            tots,
            triggereds,
            self.index)

        blob['Hits'] = hit_series

        event_info = EventInfo((
            header.det_id,
            self.index,
            # header.time_slice,
            0,  # livetime_sec
            0, 0,  # MC ID and time
            0, 0,  # n evts/files gen
            event.overlays,
            # header.run,
            event.trigger_counter, event.trigger_mask,
            header.ticks * 16, header.time_stamp,
            0, 0, 0,  # MC weights
            0))
        blob['EventInfo'] = event_info

        self.index += 1

    def process_summaryslice(self, data, blob):
        data_io = BytesIO(data)
        preamble = DAQPreamble(file_obj=data_io)
        summaryslice = DAQSummaryslice(file_obj=data_io)
        blob["RawSummaryslice"] = summaryslice


class DAQPreamble(object):
    """Wrapper for the JDAQPreamble binary format.

    Args:
      file_obj (file): The binary file, where the file pointer is at the
        beginning of the header.

    Attributes:
      size (int): The size of the original DAQ byte representation.
      data_type (int): The data type of the following frame. The coding is
        stored in the ``DATA_TYPES`` dictionary::

            101: 'DAQSuperFrame'
            201: 'DAQSummaryFrame'
            1001: 'DAQTimeslice'
            2001: 'DAQSummaryslice'
            10001: 'DAQEvent'

    """
    size = 8

    def __init__(self, byte_data=None, file_obj=None):
        self.length = None
        self.data_type = None
        if byte_data:
            self._parse_byte_data(byte_data)
        if file_obj:
            self._parse_file(file_obj)

    def _parse_byte_data(self, byte_data):
        """Extract the values from byte string."""
        self.length, self.data_type = unpack('<ii', byte_data[:self.size])

    def _parse_file(self, file_obj):
        """Directly read from file handler.

        Note that this will move the file pointer.

        """
        byte_data = file_obj.read(self.size)
        self._parse_byte_data(byte_data)

    def __repr__(self):
        description = "Length: {0}\nDataType: {1}"\
            .format(self.length, self.data_type)
        return description


class DAQHeader(object):
    """Wrapper for the JDAQHeader binary format.

    Args:
      file_obj (file): The binary file, where the file pointer is at the
        beginning of the header.

    Attributes:
      size (int): The size of the original DAQ byte representation.

    """
    size = 20

    def __init__(self, byte_data=None, file_obj=None):
        self.run = None
        self.time_slice = None
        self.time_stamp = None
        if byte_data:
            self._parse_byte_data(byte_data)
        if file_obj:
            self._parse_file(file_obj)

    def _parse_byte_data(self, byte_data):
        """Extract the values from byte string."""
        chunks = unpack('<iiiii', byte_data[:self.size])
        det_id, run, time_slice, time_stamp, ticks = chunks
        self.det_id = det_id
        self.run = run
        self.time_slice = time_slice
        self.time_stamp = time_stamp
        self.ticks = ticks

    def _parse_file(self, file_obj):
        """Directly read from file handler.

        Note:
          This will move the file pointer.

        """
        byte_data = file_obj.read(self.size)
        self._parse_byte_data(byte_data)

    def __repr__(self):
        description = "Run: {0}\nTime slice: {1}\nTime stamp: {2} ({3})"\
                      .format(self.run, self.time_slice, self.time_stamp,
                              self.ticks)
        return description


class DAQSummaryslice(object):
    """Wrapper for the JDAQSummarySlice binary format.

    Args:
      file_obj (file): The binary file, where the file pointer is at the
        beginning of the header.

    Attributes:
      n_summary_frames (int): The number of summary frames.
      summary_frames (dict): The PMT rates for each DOM. The key is the DOM
        identifier and the corresponding value is a sorted list of PMT rates
        in [Hz].
      dom_rates (dict): The overall DOM rate for each DOM.

    """
    def __init__(self, file_obj):
        self.header = DAQHeader(file_obj=file_obj)
        self.n_summary_frames = unpack('<i', file_obj.read(4))[0]
        self.summary_frames = {}
        self.dq_status = {}
        self.dom_status = {}
        self.dom_rates = {}

        self._parse_summary_frames(file_obj)

    def _parse_summary_frames(self, file_obj):
        """Iterate through the byte data and fill the summary_frames"""
        for _ in range(self.n_summary_frames):
            dom_id = unpack('<i', file_obj.read(4))[0]
            dq_status = file_obj.read(4)  # probably dom status? # noqa
            dom_status = unpack('<iiii', file_obj.read(16))
            raw_rates = unpack('b'*31, file_obj.read(31))
            pmt_rates = [self._get_rate(value) for value in raw_rates]
            self.summary_frames[dom_id] = pmt_rates
            self.dq_status[dom_id] = dq_status
            self.dom_status[dom_id] = dom_status
            self.dom_rates[dom_id] = np.sum(pmt_rates)

    def _get_rate(self, value):
        """Return the rate in Hz from the short int value"""
        if value == 0:
            return 0
        else:
            return MINIMAL_RATE_HZ * math.exp(value * self._get_factor())

    def _get_factor(self):
        return math.log(MAXIMAL_RATE_HZ / MINIMAL_RATE_HZ) / 255


class DAQEvent(object):
    """Wrapper for the JDAQEvent binary format.

    Args:
      file_obj (file): The binary file, where the file pointer is at the

    Attributes:
      trigger_counter (int): Incremental identifier of the occurred trigger.
      trigger_mask (int): The trigger type(s) satisfied.
      overlays (int): Number of merged events.
      n_triggered_hits (int): Number of hits satisfying the trigger conditions.
      n_snapshot_hits (int): Number of snapshot hits.
      triggered_hits (list): A list of triggered hits
        (dom_id, pmt_id, tdc_time, tot, (trigger_mask,))
      snapshot_hits (list): A list of snapshot hits
        (dom_id, pmt_id, tdc_time, tot)

    """
    def __init__(self, file_obj):
        self.header = DAQHeader(file_obj=file_obj)
        self.trigger_counter = unpack('<Q', file_obj.read(8))[0]
        self.trigger_mask = unpack('<Q', file_obj.read(8))[0]
        self.overlays = unpack('<i', file_obj.read(4))[0]

        self.n_triggered_hits = unpack('<i', file_obj.read(4))[0]
        self.triggered_hits = []
        self._parse_triggered_hits(file_obj)

        self.n_snapshot_hits = unpack('<i', file_obj.read(4))[0]
        self.snapshot_hits = []
        self._parse_snapshot_hits(file_obj)

    def _parse_triggered_hits(self, file_obj):
        """Parse and store triggered hits."""
        for _ in range(self.n_triggered_hits):
            dom_id, pmt_id = unpack('<ib', file_obj.read(5))
            tdc_time = unpack('>I', file_obj.read(4))[0]
            tot = unpack('<b', file_obj.read(1))[0]
            trigger_mask = unpack('<Q', file_obj.read(8))
            self.triggered_hits.append((dom_id, pmt_id, tdc_time, tot,
                                       trigger_mask))

    def _parse_snapshot_hits(self, file_obj):
        """Parse and store snapshot hits."""
        for _ in range(self.n_snapshot_hits):
            dom_id, pmt_id = unpack('<ib', file_obj.read(5))
            tdc_time = unpack('>I', file_obj.read(4))[0]
            tot = unpack('<b', file_obj.read(1))[0]
            self.snapshot_hits.append((dom_id, pmt_id, tdc_time, tot))

    def __repr__(self):
        string = '\n'.join((
            " Number of triggered hits: " + str(self.n_triggered_hits),
            " Number of snapshot hits: " + str(self.n_snapshot_hits)
        ))
        string += "\nTriggered hits:\n"
        string += pprint.pformat(self.triggered_hits)
        string += "\nSnapshot hits:\n"
        string += pprint.pformat(self.snapshot_hits)
        return string


class THMCData(object):
    """Monitoring Channel data."""
    def __init__(self, file_obj):
        f = file_obj
        self.data_type = unpack('>I', f.read(4))[0]
        self.run = unpack('>I', f.read(4))[0]
        self.sequence_number = unpack('>I', f.read(4))[0]  # not sure
        self.utc_seconds = unpack('>I', f.read(4))[0]
        self.utc_nanoseconds = unpack('>I', f.read(4))[0] * 16
        self.dom_id = unpack('>I', f.read(4))[0]
        self.dom_status_1 = unpack('>I', f.read(4))[0]  # not sure
        self.dom_status_2 = unpack('>I', f.read(4))[0]  # not sure
        self.dom_status_3 = unpack('>I', f.read(4))[0]  # not sure
        self.dom_status_4 = unpack('>I', f.read(4))[0]  # not sure
        self.pmt_rates = [r*10.0 for r in unpack('>' + 31*'I', f.read(31*4))]
        self.pad = unpack('>I', f.read(4))[0]  # not sure
        self.valid = unpack('>I', f.read(4))[0]  # not sure
        self.yaw, self.pitch, self.roll = unpack('>fff', f.read(12))
        self.ax, self.ay, self.az = unpack('>fff', f.read(12))
        self.gx, self.gy, self.gz = unpack('>fff', f.read(12))
        self.hx, self.hy, self.hz = unpack('>fff', f.read(12))
        self.temp = unpack('>H', f.read(2))[0] / 100.0
        self.humidity = unpack('>H', f.read(2))[0] / 100.0
        # self.det_id = unpack('>I', f.read(4))[0]  # not sure
        # self.n_packets = unpack('>H', f.read(2))[0]  # not sure
        # self.highest_packet_number = unpack('>H', f.read(2))[0]  # not sure
        # self.n_items = unpack('>I', f.read(4))[0]  # not sure

    def __str__(self):
        return str(vars(self))

    def __repr__(self):
        return self.__str__()



# coding=utf-8
# cython: profile=True
# Filename: tools.pyx
# cython: embedsignature=True
# pylint: disable=C0103
"""
Some unsorted, frequently used logic.

"""
from __future__ import division, absolute_import, print_function

import resource
import sys
import os
import base64
import subprocess
import collections
import socket
from collections import namedtuple
from itertools import chain
from datetime import datetime
import time
from timeit import default_timer as timer
from contextlib import contextmanager
import re
import warnings

import numpy as np
import scipy.linalg


from .logger import logging

__author__ = "Tamas Gal and Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = logging.getLogger(__name__)  # pylint: disable=C0103


def unpack_nfirst(seq, nfirst):
    """Unpack the nfrist items from the list and return the rest.

    >>> a, b, c, rest = unpack_nfirst((1, 2, 3, 4, 5), 3)
    >>> a, b, c
    (1, 2, 3)
    >>> rest
    (4, 5)

    """
    iterator = iter(seq)
    for _ in range(nfirst):
        yield next(iterator, None)
    yield tuple(iterator)


def split(string, callback=None, sep=' '):
    """Split the string and execute the callback function on each part.

    >>> string = "1 2 3 4"
    >>> parts = split(string, int)
    >>> parts
    [1, 2, 3, 4]

    """
    if callback is not None:
        return [callback(i) for i in string.split(sep)]
    else:
        return string.split(sep)


def namedtuple_with_defaults(typename, field_names, default_values=[]):
    """Create a namedtuple with default values

    >>> Node = namedtuple_with_defaults('Node', 'val left right')
    >>> Node()
    Node(val=None, left=None, right=None)
    >>> Node = namedtuple_with_defaults('Node', 'val left right', [1, 2, 3])
    >>> Node()
    Node(val=1, left=2, right=3)
    >>> Node = namedtuple_with_defaults('Node', 'val left right', {'right':7})
    >>> Node()
    Node(val=None, left=None, right=7)
    >>> Node(4)
    Node(val=4, left=None, right=7)
    """
    the_tuple = namedtuple(typename, field_names)
    the_tuple.__new__.__defaults__ = (None,) * len(the_tuple._fields)
    if isinstance(default_values, collections.Mapping):
        prototype = the_tuple(**default_values)
    else:
        prototype = the_tuple(*default_values)
    the_tuple.__new__.__defaults__ = tuple(prototype)
    return the_tuple


def zenith(v):
    """Return the zenith angle in radians"""
    return angle_between((0, 0, -1), v)


def azimuth(v):
    """Return the azimuth angle in radians"""
    v = np.atleast_2d(v)
    phi = np.arctan2(v[:, 1], v[:, 0])
    phi[phi < 0] += 2 * np.pi
    if len(phi) == 1:
        return phi[0]
    return phi


def cartesian(phi, theta, radius=1):
    x = radius * np.sin(theta) * np.cos(phi)
    y = radius * np.sin(theta) * np.sin(phi)
    z = radius * np.cos(theta)
    return np.column_stack((x, y, z))


def angle_between(v1, v2):
    """Returns the angle in radians between vectors 'v1' and 'v2'::

    >>> angle_between((1, 0, 0), (0, 1, 0))
    1.5707963267948966
    >>> angle_between((1, 0, 0), (1, 0, 0))
    0.0
    >>> angle_between((1, 0, 0), (-1, 0, 0))
    3.141592653589793

    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    # Don't use `np.dot`, does not work with all shapes
    angle = np.arccos(np.inner(v1_u, v2_u))
    return angle


def unit_vector(vector, **kwargs):
    """Returns the unit vector of the vector."""
    # This also works for a dataframe with columns ['x', 'y', 'z']
    # However, the division operation is picky about the shapes
    # So, remember input vector shape, cast all up to 2d,
    # do the (ugly) conversion, then return unit in same shape as input
    # of course, the numpy-ized version of the input...
    vector = np.array(vector)
    out_shape = vector.shape
    vector = np.atleast_2d(vector)
    unit = vector / scipy.linalg.norm(vector, axis=1, **kwargs)[:, None]
    return unit.reshape(out_shape)


def pld3(p1, p2, d2):
    """Calculate the point-line-distance for given point and line."""
    return scipy.linalg.norm(np.cross(d2, p2 - p1)) / scipy.linalg.norm(d2)


def lpnorm(x, p=2):
    return np.power(np.sum(np.power(x, p)), 1/p)


def dist(x1, x2):
    return lpnorm(x2 - x1, p=2)


def com(points, masses=None):
    """Calculate center of mass for given points.
    If masses is not set, assume equal masses."""
    if masses is None:
        return np.average(points, axis=0)
    else:
        return np.average(points, axis=0, weights=masses)


def circ_permutation(items):
    """Calculate the circular permutation for a given list of items."""
    permutations = []
    for i in range(len(items)):
        permutations.append(items[i:] + items[:i])
    return permutations


def geant2pdg(geant_code):
    """Convert GEANT particle ID to PDG"""
    conversion_table = {
        1: 22,     # photon
        2: -11,    # positron
        3: 11,     # electron
        5: -13,    # muplus
        6: 13,     # muminus
        7: 111,    # pi0
        8: 211,    # piplus
        9: -211,   # piminus
        10: 130,   # k0long
        11: 321,   # kplus
        12: -321,  # kminus
        13: 2112,  # neutron
        14: 2212,  # proton
        16: 310,   # kaon0short
        17: 221,   # eta
        }
    try:
        return conversion_table[geant_code]
    except KeyError:
        return 0


_PDG2NAME = {
    1: 'd',
    2: 'u',
    3: 's',
    4: 'c',
    5: 'b',
    6: 't',
    11: 'e-',
    -11: 'e+',
    12: 'nu_e',
    -12: 'anu_e',
    13: 'mu-',
    -13: 'mu+',
    14: 'nu_mu',
    -14: 'anu_mu',
    15: 'tau-',
    -15: 'tau+',
    16: 'nu_tau',
    -16: 'anu_tau',
    22: 'photon',
    111: 'pi0',
    130: 'K0L',
    211: 'pi-',
    -211: 'pi+',
    310: 'K0S',
    311: 'K0',
    321: 'K+',
    -321: 'K-',
    2112: 'n',
    2212: 'p',
    -2212: 'p-',
}

_NAME2PDG = {val: key for key, val in _PDG2NAME.items()}      # noqa


def pdg2name(pdg_id):
    """Convert PDG ID to human readable names"""
    # pylint: disable=C0330
    try:
        return _PDG2NAME[pdg_id]
    except KeyError:
        return "N/A"

def name2pdg(name):
    try:
        return _NAME2PDG[name]
    except KeyError:
        return 0


def total_seconds(td):
    """Convert the timedelta to seconds. (Python 2.6 backward compatibility)"""
    try:
        s = td.total_seconds()
    except AttributeError:
        s = (td.microseconds + (td.seconds + td.days*24*3600) * 10**6) / 10**6
    return s


class Timer(object):
    """A very simple, accurate and easy to use timer context"""
    def __init__(self, message='It', precision=3, callback=print):
        self.message = message
        self.precision = precision
        self.callback = callback

    def __enter__(self):
        self.start()

    def __exit__(self, type, value, traceback):
        self.stop()

    def start(self):
        self.__start = timer()
        self.__start_cpu = time.clock()

    def stop(self):
        self.__finish = timer()
        self.__finish_cpu = time.clock()
        self.log()
        return self.seconds

    @property
    def seconds(self):
        return self.__finish - self.__start

    @property
    def cpu_seconds(self):
        return self.__finish_cpu - self.__start_cpu

    def log(self):
        self.callback("{0} took {1:.{3}f}s (CPU {2:.{3}f}s)."
                      .format(self.message,
                              self.seconds, self.cpu_seconds,
                              self.precision))


class Cuckoo(object):
    "A timed callback caller, which only executes once in a given interval."
    def __init__(self, interval=0, callback=print):
        "Setup with interval in seconds and a callback function"
        self.interval = interval
        self.callback = callback
        self.timestamp = None

    def msg(self, *args, **kwargs):
        "Only execute callback when interval is reached."
        if self.timestamp is None or self._interval_reached():
            self.callback(*args, **kwargs)
            self.reset()

    def reset(self):
        "Reset the timestamp"
        self.timestamp = datetime.now()

    def _interval_reached(self):
        "Check if defined interval is reached"
        return total_seconds(datetime.now() - self.timestamp) > self.interval

    def __call__(self, *args, **kwargs):
        "Run the msg function when called directly."
        self.msg(*args, **kwargs)


def ifiles(irods_path):
    """Return a list of filenames for given iRODS path (recursively)"""
    raw_output = subprocess.check_output("ils -r --bundle {0}"
                                         "    | grep 'Bundle file:'"
                                         "    | awk '{{print $3}}'"
                                         .format(irods_path), shell=True)
    filenames = raw_output.strip().split("\n")
    return filenames


def remain_file_pointer(function):
    """Remain the file pointer position after calling the decorated function

    This decorator assumes that the last argument is the file handler.

    """
    def wrapper(*args, **kwargs):
        """Wrap the function and remain its parameters and return values"""
        file_obj = args[-1]
        old_position = file_obj.tell()
        return_value = function(*args, **kwargs)
        file_obj.seek(old_position, 0)
        return return_value
    return wrapper


def peak_memory_usage():
    """Return peak memory usage in MB"""
    mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    factor_mb = 1 / 1024
    if sys.platform == 'darwin':
        factor_mb = 1 / (1024 * 1024)
    return mem * factor_mb


@contextmanager
def ignored(*exceptions):
    """Ignore-context for a given list of exceptions.

    Example:
        with ignored(AttributeError):
            foo.a = 1

    """
    try:
        yield
    except exceptions:
        pass


def token_urlsafe(nbytes=32):
    """Return a random URL-safe text string, in Base64 encoding.

    This is taken and slightly modified from the Python 3.6 stdlib.

    The string has *nbytes* random bytes.  If *nbytes* is ``None``
    or not supplied, a reasonable default is used.

    >>> token_urlsafe(16)  #doctest:+SKIP
    'Drmhze6EPcv0fN_81Bj-nA'

    """
    tok = os.urandom(nbytes)
    return base64.urlsafe_b64encode(tok).rstrip(b'=').decode('ascii')


def tai_timestamp():
    """Return current TAI timestamp."""
    timestamp = time.time()
    date = datetime.utcfromtimestamp(timestamp)
    if date.year < 1972:
        return timestamp
    offset = 10 + timestamp
    leap_seconds = [
        (1972, 1, 1), (1972, 7, 1), (1973, 1, 1),
        (1974, 1, 1), (1975, 1, 1), (1976, 1, 1),
        (1977, 1, 1), (1978, 1, 1), (1979, 1, 1),
        (1980, 1, 1), (1981, 7, 1), (1982, 7, 1),
        (1983, 7, 1), (1985, 7, 1), (1988, 1, 1),
        (1990, 1, 1), (1991, 1, 1), (1992, 7, 1),
        (1993, 7, 1), (1994, 7, 1), (1996, 1, 1),
        (1997, 7, 1), (1999, 1, 1), (2006, 1, 1),
        (2009, 1, 1), (2012, 7, 1), (2015, 7, 1),
        (2017, 1, 1),
    ]
    for idx, leap_date in enumerate(leap_seconds):
        if leap_date >= (date.year, date.month, date.day):
            return idx - 1 + offset
    return len(leap_seconds) - 1 + offset


try:
    dict.iteritems
except AttributeError:
    # for Python 3

    def itervalues(d):
        return iter(d.values())

    def iteritems(d):
        return iter(d.items())
else:
    # for Python 2
    def itervalues(d):
        return d.itervalues()

    def iteritems(d):
        return d.iteritems()


def decamelise(text):
    """Convert CamelCase to lower_and_underscore."""
    s = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', text)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s).lower()


def camelise(text, capital_first=True):
    """Convert lower_underscore to CamelCase."""
    def camelcase():
        if not capital_first:
            yield str.lower
        while True:
            yield str.capitalize

    c = camelcase()
    return "".join(next(c)(x) if x else '_' for x in text.split("_"))


def insert_prefix_to_dtype(arr, prefix):
    new_cols = [prefix + '_' + col for col in arr.dtype.names]
    arr.dtype.names = new_cols
    return arr


class deprecated(object):
    """Decorator to mark a function or class as deprecated.

    >>> @deprecated('some warning')
    ... def some_function(): pass
    """

    # Adapted from http://wiki.python.org/moin/PythonDecoratorLibrary,
    # but with many changes.
    # and stolen again from sklearn.utils

    def __init__(self, extra=''):
        """
        Parameters
        ----------
        extra: string
          to be added to the deprecation messages
        """
        self.extra = extra

    def __call__(self, obj):
        if isinstance(obj, type):
            return self._decorate_class(obj)
        else:
            return self._decorate_fun(obj)

    def _decorate_class(self, cls):
        msg = "Class %s is deprecated" % cls.__name__
        if self.extra:
            msg += "; %s" % self.extra

        init = cls.__init__

        def wrapped(*args, **kwargs):
            warnings.warn(msg, category=DeprecationWarning)
            return init(*args, **kwargs)
        cls.__init__ = wrapped

        wrapped.__name__ = '__init__'
        wrapped.__doc__ = self._update_doc(init.__doc__)
        wrapped.deprecated_original = init

        return cls

    def _decorate_fun(self, fun):
        """Decorate function fun"""

        msg = "Function %s is deprecated" % fun.__name__
        if self.extra:
            msg += "; %s" % self.extra

        def wrapped(*args, **kwargs):
            warnings.warn(msg, category=DeprecationWarning)
            return fun(*args, **kwargs)

        wrapped.__name__ = fun.__name__
        wrapped.__dict__ = fun.__dict__
        wrapped.__doc__ = self._update_doc(fun.__doc__)

        return wrapped

    def _update_doc(self, olddoc):
        newdoc = "DEPRECATED"
        if self.extra:
            newdoc = "%s: %s" % (newdoc, self.extra)
        if olddoc:
            newdoc = "%s\n\n%s" % (newdoc, olddoc)
        return newdoc


def add_empty_flow_bins(bins):
    """Add empty over- and underflow bins.
    """
    bins = list(bins)
    bins.insert(0, 0)
    bins.append(0)
    return np.array(bins)


def flat_weights(x, bins):
    """Get weights to produce a flat histogram.
    """
    bin_width = np.abs(bins[1] - bins[0])
    hist, _ = np.histogram(x, bins=bins)
    hist = hist.astype(float)
    hist = add_empty_flow_bins(hist)
    hist *= bin_width
    which = np.digitize(x, bins=bins, right=True)
    pop = hist[which]
    wgt = 1 / pop
    wgt *= len(wgt) / np.sum(wgt)
    return wgt


def prettyln(text, fill='-', align='^', prefix='[ ', suffix=' ]', length=69):
    """Wrap `text` in a pretty line with maximum length."""
    text = '{prefix}{0}{suffix}'.format(text, prefix=prefix, suffix=suffix)
    print("{0:{fill}{align}{length}}"
          .format(text, fill=fill, align=align, length=length))


def irods_filepath(det_id, run_id):
    """Generate the iRODS filepath for given detector (O)ID and run ID"""
    data_path = "/in2p3/km3net/data/raw/sea"
    from km3pipe.db import DBManager
    if not isinstance(det_id, int):
        dts = DBManager().detectors
        det_id = int(dts[dts.OID == det_id].SERIALNUMBER.values[0])
    return data_path + "/KM3NeT_{0:08}/{2}/KM3NeT_{0:08}_{1:08}.root" \
           .format(det_id, run_id, run_id//1000)



# Filename: aanet.py
# pylint: disable=locally-disabled
"""
Pump for the Aanet data format.

This is undoubtedly the ugliest module in the entire framework.
If you have a way to read aanet files via the Jpp interface,
your pull request is more than welcome!
"""
from __future__ import division, absolute_import, print_function
import os.path

import numpy as np

from km3pipe.core import Pump, Blob
from km3pipe.dataclasses import (HitSeries, TrackSeries, EventInfo,
                                 KM3Array, KM3DataFrame)
from km3pipe.logger import logging

log = logging.getLogger(__name__)  # pylint: disable=C0103

__author__ = "Tamas Gal, Thomas Heid and Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = ["Liam Quin & Javier Barrios Marti"]
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class AanetPump(Pump):
    """A pump for binary Aanet files.

    Parameters
    ----------
    filename: str, optional
        Name of the file to open. If this parameter is not given, ``filenames``
        needs to be specified instead.
    filenames: list(str), optional
        List of files to open.
    aa_fmt: string, optional
        Subformat of aanet in the file. Possible values:
        ``'minidst', 'jevt_jgandalf', 'generic_track', 'ancient_recolns'``
    use_aart_sps_lib: bool, optional [default=False]
        Use Aart's local SPS copy (@ Lyon) via ``gSystem.Load``?
    """

    def __init__(self, **context):
        super(self.__class__, self).__init__(**context)

        self.filename = self.get('filename')
        self.filenames = self.get('filenames') or []
        self.indices = self.get('indices')
        self.additional = self.get('additional')
        self.format = self.get('aa_fmt')
        self.use_id = self.get('use_id') or False
        self.use_aart_sps_lib = self.get('use_aart_sps_lib') or False
        if self.additional:
            self.id = self.get('id')
            self.return_without_match = self.get("return_without_match")

        if not self.filename and not self.filenames:
            raise ValueError("No filename(s) defined")

        if self.additional and not self.filename:
            log.error("additional file only implemeted for single files")

        if self.filename:
            if "[index]" in self.filename and self.indices:
                self._parse_filenames()
            else:
                self.filenames.append(self.filename)

        import ROOT # noqa
        self.aalib = self.get('aa_lib')
        if self.aalib:
            ROOT.gSystem.Load(self.aalib) # noqa
        elif self.use_aart_sps_lib:
            aapath = '/sps/km3net/users/heijboer/aanet/libaa.so'
            ROOT.gSystem.Load(aapath) # noqa
        else:
            import aa  # noqa

        self.header = None
        self.blobs = self.blob_generator()
        self.i = 0

        if self.additional:
            dummy_evt = ROOT.Evt()
            dummy_evt.frame_index = -1
            self.previous = {"Evt": dummy_evt}

    def _parse_filenames(self):
        prefix, suffix = self.filename.split('[index]')
        self.filenames += [prefix + str(i) + suffix for i in self.indices]

    def get_blob(self, index):
        NotImplementedError("Aanet currently does not support indexing.")

    def blob_generator(self):
        """Create a blob generator."""
        # pylint: disable:F0401,W0612
        from ROOT import EventFile

        found_any_files = False
        for filename in self.filenames:
            print("Reading from file: {0}".format(filename))
            if not os.path.exists(filename):
                log.warn(filename + " not available: continue without it")
                continue

            if not found_any_files:
                found_any_files = True

            try:
                event_file = EventFile(filename)
            except Exception:
                raise SystemExit("Could not open file")

            # http://jenkins.km3net.de/job/aanet_trunk/doxygen/Head_8hh_source.html
            livetime = 0
            livetime_err = 0
            n_gen = 0
            nfilgen = 0

            try:
                self.header = event_file.rootfile().Get("Head")
            except (ValueError, UnicodeEncodeError):
                log.warn(filename + ": can't read header.")
            try:
                lt_line = self.header.get_line('livetime')
                if len(lt_line) > 0:
                    livetime, livetime_err = lt_line.split()
                #livetime = self.header.get_field('livetime', 0)
                #livetime_err = self.header.get_field('livetime', 1)
                self.livetime_err = float(livetime_err)
                self.livetime = float(livetime)
            except (ValueError, UnicodeEncodeError):
                log.warn(filename + ": can't read livetime.")
                self.livetime = 0
            try:
                ngen = self.header.get_field('genvol', 4)
                self.ngen = float(ngen)
            except (ValueError, UnicodeEncodeError):
                log.warn(filename + ": can't read ngen.")
                self.ngen = 0
            try:
                self.nfilgen = float(nfilgen)
            except (ValueError, UnicodeEncodeError):
                log.warn(filename + ": can't read nfilgen.")
                self.nfilgen = 0

            if self.format == 'ancient_recolns':
                while event_file.next():
                    event = event_file.evt
                    blob = self._read_event(event, filename)
                    yield blob
            else:
                for event in event_file:
                    blob = self._read_event(event, filename)
                    yield blob
            del event_file

        if not found_any_files:
            msg = "None of the input files available. Exiting."
            try:
                raise FileNotFoundError(msg)
            except NameError:
                # python2
                raise IOError(msg)

    def _read_event(self, event, filename):
        try:
            if event.det_id <= 0:  # apply ZED correction
                for track in event.mc_trks:
                    track.pos.z += 405.93
        except AttributeError:
            pass

        if len(event.w) == 3:
            w1, w2, w3 = event.w
        else:
            w1 = w2 = w3 = np.nan

        blob = Blob()
        blob['Evt'] = event
        if self.use_id:
            event_id = event.id
        else:
            event_id = self.i
        self.i += 1
        try:
            blob['Hits'] = HitSeries.from_aanet(event.hits, event_id)
            blob['McHits'] = HitSeries.from_aanet(event.mc_hits,
                                                  event_id)
        except AttributeError:
            pass
        blob['McTracks'] = TrackSeries.from_aanet(event.mc_trks,
                                                  event_id)
        blob['filename'] = filename
        try:
            blob['EventInfo'] = EventInfo((
                event.det_id,
                event.frame_index,
                self.livetime, # livetime_sec
                event.mc_id,
                event.mc_t,
                self.ngen,   # n_events_gen
                self.nfilgen, # n_files_gen
                event.overlays,
                # event.run_id,
                event.trigger_counter,
                event.trigger_mask,
                event.t.GetNanoSec(),
                event.t.GetSec(),
                w1, w2, w3,
                event_id))
        except AttributeError:
            blob['EventInfo'] = EventInfo((0,   # det_id
                                           event.frame_index,
                                           self.livetime,   # livetime_sec
                                           0,   # mc_id
                                           0,   # mc_t
                                           self.ngen,   # n_events_gen
                                           self.nfilgen,   # n_files_gen
                                           0,   # overlays
                                           0,   # trigger_counter
                                           0,   # trigger_mask
                                           0,   # nanose
                                           0,   # sec
                                           w1, w2, w3,
                                           event_id))
        if self.format == 'minidst':
            recos = read_mini_dst(event, event_id)
            for recname, reco in recos.items():
                blob[recname] = reco
        if self.format in ('jevt_jgandalf', 'gandalf', 'jgandalf'):
            track, dtype = parse_jevt_jgandalf(event, event_id)
            if track:
                blob['Gandalf'] = KM3Array.from_dict(track, dtype,
                                                           h5loc='/reco')
        if self.format == 'generic_track':
            track, dtype = parse_generic_event(event, event_id)
            if track:
                blob['Track'] = KM3Array.from_dict(track, dtype,
                                                    h5loc='/reco')
        if self.format in ('ancient_recolns', 'orca_recolns'):
            track, dtype = parse_ancient_recolns(event, event_id)
            if track:
                blob['OrcaRecoLns'] = KM3Array.from_dict(track, dtype,
                                                             h5loc='/reco')
        return blob

    def event_index(self, blob):
        if self.id:
            return blob["Evt"].id
        else:
            return blob["Evt"].frame_index

    def process(self, blob=None):
        if self.additional:
            new_blob = self.previous
            if self.event_index(new_blob) == blob["Evt"].frame_index:
                blob[self.additional] = new_blob
                return blob

            elif self.event_index(new_blob) > blob["Evt"].frame_index:
                if self.return_without_match:
                    return blob
                else:
                    return None

            else:
                while self.event_index(new_blob) < blob["Evt"].frame_index:
                    new_blob = next(self.blobs)

                self.previous = new_blob

                return self.process(blob)

        else:
            return next(self.blobs)

    def __iter__(self):
        return self

    def next(self):
        """Python 2/3 compatibility for iterators"""
        return self.__next__()

    def __next__(self):
        return next(self.blobs)


def parse_ancient_recolns(aanet_event, event_id):
    # the final recontructed track is in f.evt.trks, at the index 5
    # trks[0] ==> prefit track
    # trks[1] ==> m-estimator track
    # trks[2] ==> "original" pdf track
    # trks[3] ==> full pdf track, final reconstructed direction
    # trks[4] ==> + vertex fit and muon length fit
    # trks[5] ==> + bjorken-y fit and neutrino energy fit
    try:
        out = {}
        out['quality'] = aanet_event.trks[3].lik
        out['n_hits_used'] = aanet_event.trks[3].fitinf[0]
        out['energy_muon'] = aanet_event.trks[4].E
        out['energy_neutrino'] = aanet_event.trks[5].E
        out['bjorken_y'] = aanet_event.trks[5].By_rec
        out['pos_x'] = aanet_event.trks[5].pos.x
        out['pos_y'] = aanet_event.trks[5].pos.y
        out['pos_z'] = aanet_event.trks[5].pos.z
        out['dir_x'] = aanet_event.trks[5].dir.x
        out['dir_y'] = aanet_event.trks[5].dir.y
        out['dir_z'] = aanet_event.trks[5].dir.z

        sigma2_theta = aanet_event.trks[3].error_matrix[18] \
            if aanet_event.trks[3].error_matrix[18] > 0 else 0
        out['sigma2_theta'] = sigma2_theta
        sigma2_phi = aanet_event.trks[3].error_matrix[24] \
            if aanet_event.trks[3].error_matrix[24] > 0 else 0
        out['sigma2_phi'] = sigma2_phi
        sin_theta = np.sin(np.arccos(aanet_event.trks[3].dir.z))
        out['sin_theta'] = sin_theta
        out['beta'] = np.sqrt(sin_theta * sin_theta * sigma2_phi + sigma2_theta)
    except IndexError:
        keys = {'quality', 'n_hits_used', 'pos_x', 'pos_y', 'pos_z',        # noqa
                'dir_x', 'dir_y', 'dir_z', 'energy_muon', 'energy_neutrino',        # noqa
                'bjorken_y', 'beta', 'sigma2_theta', 'sigma2_phi',      # noqa
                'sin_theta', }    #noqa
        out = {key: 0 for key in keys}
    dt = [(key, float) for key in sorted(out.keys())]
    out['event_id'] = event_id
    dt.append(('event_id', '<u4'))
    dt = np.dtype(dt)
    return out, dt


def parse_jevt_jgandalf(aanet_event, event_id):
    try:
        track = aanet_event.trks[0]     # this might throw IndexError
        map = {}
        map['id'] = track.id
        map['pos_x'] = track.pos.x
        map['pos_y'] = track.pos.y
        map['pos_z'] = track.pos.z
        map['dir_x'] = track.dir.x
        map['dir_y'] = track.dir.y
        map['dir_z'] = track.dir.z
        map['time'] = track.t
        map['type'] = track.type
        map['rec_type'] = track.rec_type
        map['rec_stage'] = track.rec_stage
        map['beta0'] = track.fitinf[0]
        map['beta1'] = track.fitinf[1]
        map['lik'] = track.fitinf[2]
        map['lik_red'] = track.fitinf[3]
        map['energy'] = track.fitinf[4]
    except IndexError:
        keys = {'id', 'pos_x', 'pos_y', 'pos_z', 'dir_x', 'dir_y', 'dir_z',
                'time', 'type', 'rec_type', 'rec_stage', 'beta0', 'beta1',
                'lik', 'lik_red', 'energy', }
        map = {key: 0 for key in keys}
    dt = [(key, float) for key in sorted(map.keys())]
    map['event_id'] = event_id
    dt.append(('event_id', '<u4'))
    dt = np.dtype(dt)
    return map, dt


def parse_generic_event(aanet_event, event_id):
    map = {}
    try:
        track = aanet_event.trks[0]
    except IndexError:
        #TODO: don't return empty map
        return map, None
    map['id'] = track.id
    map['pos_x'] = track.pos[0]
    map['pos_y'] = track.pos[1]
    map['pos_z'] = track.pos[2]
    map['dir_x'] = track.dir[0]
    map['dir_y'] = track.dir[1]
    map['dir_z'] = track.dir[2]
    map['energy'] = track.E
    map['time'] = track.t
    map['lik'] = track.lik
    map['type'] = track.type
    map['rec_type'] = track.rec_type
    map['rec_stage'] = track.rec_stage
    for k, entry in enumerate(track.fitinf):
        map['fitinf_{}'.format(k)] = entry
    for k, entry in enumerate(track.error_matrix):
        map['error_matrix_{}'.format(k)] = entry
    dt = [(key, float) for key in sorted(map.keys())]
    map['event_id'] = event_id
    dt.append(('event_id', '<u4'))
    dt = np.dtype(dt)
    return map, dt


def read_mini_dst(aanet_event, event_id):
    pos_to_recname = {
        0: 'RecoLNS',
        1: 'JGandalf',
        2: 'AaShowerFit',
        3: 'QStrategy',
        4: 'Dusj',
    }
    recname_to_reader = {
        'RecoLNS': parse_recolns,
        'JGandalf': parse_jgandalf,
        'AaShowerFit': parse_aashowerfit,
        'QStrategy': parse_qstrategy,
        'Dusj': parse_dusj,
    }
    minidst = {}
    if len(aanet_event.trks) == 0:
        return minidst
    for k, trk in enumerate(aanet_event.trks):
        recname = pos_to_recname[k]
        reader = recname_to_reader[recname]

        reco_map, dtype = reader(trk)
        minidst[recname] = KM3Array.from_dict(reco_map, dtype,
                                               h5loc='/reco')

    thomas_map, dtype = parse_thomasfeatures(aanet_event.usr)
    minidst['ThomasFeatures'] = KM3Array.from_dict(thomas_map, dtype,
                                                    h5loc='/reco')

    return minidst


def parse_track(trk):
    out = {}
    out['pos_x'] = trk.pos.x
    out['pos_y'] = trk.pos.y
    out['pos_z'] = trk.pos.z
    out['dir_x'] = trk.dir.x
    out['dir_y'] = trk.dir.y
    out['dir_z'] = trk.dir.z
    out['time'] = trk.t
    out['energy'] = trk.E
    out['quality'] = trk.lik
    return out


def parse_thomasfeatures(aanet_usr, event_id=0):
    out = {}
    did_converge = len(aanet_usr) > 1

    Thomas_keys = ['Slopeo1000_o100',
                   'Slope50_1000',
                   'YIntersept0_1000',
                   'Slopeo1000_1000',
                   'Chi2200_1000',
                   'TimeResidualNumberOfHits',
                   'Chi2o1000_o20',
                   'YInterspotDiffo1000_o100',
                   'YIntersepto1000_o100',
                   'Slope200_1000',
                   'Chi2o1000_0',
                   'Slope100_1000',
                   'Slopeo1000_o20',
                   'YIntersepto1000_o10',
                   'Chi2100_1000',
                   'Chi2o1000_o100',
                   'YIntersepto1000_o50',
                   'Chi2o1000_1000',
                   'YIntersept25_1000',
                   'Chi225_1000',
                   'YInterspotDiffo1000_o20',
                   'YInterspotDiffo1000_o10',
                   'Chi2o1000_o10',
                   'YInterspotDiffo1000_o50',
                   'YIntersept50_1000',
                   'TimeResidualMean',
                   'Chi20_1000',
                   'Chi2o1000_o50',
                   'YIntersept200_1000',
                   'Slope0_1000',
                   'YIntersepto1000_1000',
                   'Slope25_1000',
                   'Slopeo1000_o10',
                   'TimeResidualMedian',
                   'YIntersepto1000_o20',
                   'Slopeo1000_o50',
                   'TimeResidualRMS',
                   'YInterspotDiffo1000_0',
                   'Slopeo1000_0',
                   'YIntersepto1000_0',
                   'YIntersept100_1000',
                   'Chi250_1000',
                   'TimeResidualWidth15_85',
                   'MiddleInertia',
                   'GParameter',
                   'SmallInertia',
                   'RelativeInertia',
                   'BigInertia',
                   'GoldParameter']

    dtype = [(key, float) for key in Thomas_keys + list(out.keys())]

    out['did_converge'] = did_converge
    dtype.append(('did_converge', bool))
    out['event_id'] = event_id
    dtype.append(('event_id', '<u4'))
    dtype = np.dtype(dtype)

    if not did_converge:
        for key in Thomas_keys:
            out[key] = np.nan
    else:
        for count, key in enumerate(Thomas_keys):
            out[key] = aanet_usr[count]
    return out, dtype


def parse_recolns(aanet_trk, event_id=0):
    out = parse_track(aanet_trk)
    did_converge = aanet_trk.rec_stage > -9999

    recolns_keys = ['beta', 'n_fits', 'Lambda',
                    'n_compatible_solutions', 'Nhits', 'NhitsL0', 'NhitsL1']
    dtype = [(key, float) for key in recolns_keys + list(out.keys())]
    out['did_converge'] = did_converge
    dtype.append(('did_converge', bool))
    out['event_id'] = event_id
    dtype.append(('event_id', '<u4'))
    dtype = np.dtype(dtype)

    if not did_converge:
        for key in recolns_keys:
            out[key] = np.nan
    else:
        for count, key in enumerate(recolns_keys):
            out[key] = aanet_trk.usr[count]
    return out, dtype


def parse_jgandalf(aanet_trk, event_id=0):
    out = parse_track(aanet_trk)
    did_converge = aanet_trk.rec_stage > -9999

    jgandalf_keys = ['Energy_f', 'Energy_can', 'Beta0',
                     'Beta1', 'Lik', 'Lik_reduced', 'NhitsL0', 'NhitsL1']
    dtype = [(key, float) for key in jgandalf_keys + list(out.keys())]
    out['did_converge'] = did_converge
    dtype.append(('did_converge', bool))
    out['event_id'] = event_id
    dtype.append(('event_id', '<u4'))
    dtype = np.dtype(dtype)

    if not did_converge:
        for key in jgandalf_keys:
            out[key] = np.nan
    else:
        for count, key in enumerate(jgandalf_keys):
            out[key] = aanet_trk.usr[count]
    return out, dtype


def parse_aashowerfit(aanet_trk, event_id=0):
    out = parse_track(aanet_trk)
    did_converge = aanet_trk.rec_stage > -9999

    aashow_keys = ['NhitsAA', 'M_estimator', 'beta',
                   'NhitsL0', 'NhitsL1']
    dtype = [(key, float) for key in aashow_keys + list(out.keys())]
    out['did_converge'] = did_converge
    dtype.append(('did_converge', bool))
    out['event_id'] = event_id
    dtype.append(('event_id', '<u4'))
    dtype = np.dtype(dtype)

    if not did_converge:
        for key in aashow_keys:
            out[key] = np.nan
    else:
        for count, key in enumerate(aashow_keys):
            out[key] = aanet_trk.usr[count]
    return out, dtype


def parse_qstrategy(aanet_trk, event_id=0):
    out = parse_track(aanet_trk)
    did_converge = aanet_trk.rec_stage > -9999

    qstrat_keys = sorted(['MFinal', 'Charge', 'MPreFit',
                          'RPreFit', 'Inertia', 'NhitsL0', 'NhitsL1'])
    dtype = [(key, float) for key in qstrat_keys + list(out.keys())]
    out['did_converge'] = did_converge
    dtype.append(('did_converge', bool))
    out['event_id'] = event_id
    dtype.append(('event_id', '<u4'))
    dtype = np.dtype(dtype)

    if not did_converge:
        for key in qstrat_keys:
            out[key] = np.nan
    else:
        for count, key in enumerate(qstrat_keys):
            out[key] = aanet_trk.usr[count]
    return out, dtype


def parse_dusj(aanet_trk, event_id=0):
    out = parse_track(aanet_trk)
    did_converge = aanet_trk.rec_stage > -9999

    dusj_keys = ['BigInertia', 'Chi2100_1000', 'Chi2o1000_1000',
                 'Chi2o1000_o10', 'Chi2o1000_o100', 'Chi2o1000_o50',
                 'DusjShowerRecoFinalFitMinimizerCalls',
                 'DusjShowerRecoFinalFitReducedLogLikelihood',
                 'DusjShowerRecoVertexFitLogLikelihood',
                 'DusjShowerRecoVertexFitMinimizerCalls',
                 'DusjShowerRecoVertexFitReducedLogLikelihood',
                 'FitConvergencePositionAzimuth',
                 'FitConvergencePositionTime',
                 'FitConvergencePositionX',
                 'FitConvergencePositionY',
                 'FitConvergencePositionZ',
                 'FitConvergencePositionZenith',
                 'FitVerticalDistanceToDetectorCenter',
                 'GParameter',
                 'GoldParameter',
                 'ShowerIdentifierHorizontalDistanceToDetectorCenter',
                 'ShowerIdentifierReducedChiSquare',
                 'ShowerIdentifierVerticalDistanceToDetectorCenter',
                 'Slopeo1000_o50',
                 'SmallInertia',
                 'TimeResidualMean',
                 'TimeResidualNumberOfHits',
                 'TimeResidualRMS',
                 'YIntersept25_1000',
                 'YIntersepto1000_1000',
                 'YInterspotDiffo1000_o50',
                 'NhitsL0',
                 'NhitsL1']
    dtype = [(key, float) for key in dusj_keys + list(out.keys())]
    out['did_converge'] = did_converge
    dtype.append(('did_converge', bool))
    out['event_id'] = event_id
    dtype.append(('event_id', '<u4'))
    dtype = np.dtype(dtype)

    if not did_converge:
        for key in dusj_keys:
            out[key] = np.nan
    else:
        for count, key in enumerate(dusj_keys):
            out[key] = aanet_trk.usr[count]
    return out, dtype



#!/usr/bin/env python

import numpy as np
import tables as tb

from km3pipe.io import read_group
from km3pipe.io.pandas import H5Chain
from km3pipe.tools import insert_prefix_to_dtype
from km3pipe.testing import TestCase


class TestMultiTable(TestCase):
    def setUp(self):
        self.foo = np.array([
            (1.0, 2.0, 3.0),
            (4.0, 5.0, 6.0),
        ], dtype=[('a', '<f8'), ('b', '<f8'), ('c', '<f8'), ])
        self.bar = np.array([
            (10.0, 20.0, 30.0),
            (40.0, 50.0, 60.0),
        ], dtype=[('aa', '<f8'), ('bb', '<f8'), ('cc', '<f8'), ])
        self.tabs = {'foo': self.foo, 'bar': self.bar}
        self.where = '/lala'
        self.h5name = './test.h5'
        self.h5file = tb.open_file(
            # create the file in memory only
            self.h5name, 'w', driver="H5FD_CORE", driver_core_backing_store=0)
        for name, tab in self.tabs.items():
            self.h5file.create_table(self.where, name=name, obj=tab,
                                     createparents=True)

    def tearDown(self):
        self.h5file.close()

    def test_name_insert(self):
        exp_foo = ('foo_a', 'foo_b', 'foo_c')
        exp_bar = ('bar_aa', 'bar_bb', 'bar_cc')
        pref_foo = insert_prefix_to_dtype(self.tabs['foo'], 'foo')
        pref_bar = insert_prefix_to_dtype(self.tabs['bar'], 'bar')
        self.assertEqual(exp_foo, pref_foo.dtype.names)
        self.assertEqual(exp_bar, pref_bar.dtype.names)

    #def test_group_read(self):
    #    tabs = read_group(self.h5file.root)
    #    exp_cols = (
    #        'bar_aa', 'bar_bb', 'bar_cc',
    #        'foo_a', 'foo_b', 'foo_c',
    #    )
    #    exp_shape = (2, 6)
    #    res_shape = tabs.shape
    #    res_cols = tuple(tabs.columns)
    #    print(exp_cols)
    #    print(res_cols)
    #    self.assertEqual(exp_shape, res_shape)
    #    self.assertEqual(exp_cols, res_cols)


class TestH5Pump(TestCase):
    pass


class TestH5Sink(TestCase):
    def test_to_array(self):
        # check if is converted to array:
        # hitseries
        # reco
        pass



# coding=utf-8
# Filename: __init__.py
"""
Common unit testing support for km3pipe.

"""
from __future__ import division, absolute_import, print_function

from km3pipe.common import StringIO  # noqa
from io import BytesIO  # noqa

try:
    from unittest2 import TestCase, skip, skipIf
except ImportError:
    from unittest import TestCase, skip, skipIf  # noqa

try:
    from mock import MagicMock
    from mock import Mock
    from mock import patch
except ImportError:
    from unittest.mock import MagicMock  # noqa
    from unittest.mock import Mock  # noqa
    from unittest.mock import patch  # noqa

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"



# coding=utf-8
# Filename: cmd.py
"""
Convert HDF5 to vanilla ROOT.

Usage:
     hdf2root FILES...
     hdf2root (-h | --help)
     hdf2root --version

Options:
    -h --help           Show this screen.
"""

from __future__ import division, absolute_import, print_function

import sys
import os
from datetime import datetime

from km3pipe import version
from km3modules import StatusBar

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def hdf2root(infile, outfile):
    from rootpy.io import root_open
    from rootpy import asrootpy
    from root_numpy import array2tree
    from tables import open_file

    h5 = open_file(infile, 'r')
    rf = root_open(outfile, 'recreate')

    # 'walk_nodes' does not allow to check if is a group or leaf
    #   exception handling is bugged
    #   introspection/typecheck is buged
    # => this moronic nested loop instead of simple `walk`
    for group in h5.walk_groups():
        for leafname, leaf in group._v_leaves.items():
            tree = asrootpy(array2tree(leaf[:], name=leaf._v_pathname))
            tree.write()
    rf.close()
    h5.close()


def main():
    from docopt import docopt
    args = docopt(__doc__, version=version)

    files = args['FILES']
    for fn in files:
        hdf2root(fn, fn+'.root')



# coding=utf-8
# cython: profile=True
# Filename: dataclasses.py
# cython: embedsignature=True
# pylint: disable=W0232,C0103,C0111
"""
...

"""
from __future__ import division, absolute_import, print_function

import ctypes
from libcpp cimport bool as c_bool  # noqa
from six import with_metaclass
from struct import Struct, calcsize

import numpy as np
cimport numpy as np
cimport cython
import pandas as pd

np.import_array()

from .tools import angle_between, geant2pdg, pdg2name

__author__ = "Tamas Gal and Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"
__all__ = ('EventInfo', 'Point', 'Position', 'Direction', 'HitSeries',
           'TimesliceHitSeries', 'Hit',
           'Track', 'TrackSeries', 'Serialisable', 'SummaryframeInfo',
           'BinaryStruct')


IS_CC = {
    3: 0,         # False,
    2: 1,         # True,
    1: 0,         # False,
    0: 1,         # True,
}


class Serialisable(type):
    """A metaclass for serialisable classes.

    The classes should define a `dtype` attribute in their body and are not
    meant to define `__init__` (it will be overwritten).

    Example using six.with_metaclass for py2/py3 compat
    ---------------------------------------------------

        class Foo(with_metaclass(Serialisable)):
            dtype = np.dtype([('a', '<i4'), ('b', '>i8')])

    """
    def __new__(metaclass, class_name, class_parents, class_attr):
        attr = {'h5loc': '/'}
        for name, val in class_attr.items():
            if name == 'dtype':
                attr['dtype'] = np.dtype(val)
            else:
                attr[name] = val

        def __init__(self, *args, **kwargs):
            """Take care of the attribute settings."""
            for arg, name in zip(args, self.dtype.names):
                setattr(self, name, arg)
            for key, value in kwargs.iteritems():
                setattr(self, key, value)

        attr['__init__'] = __init__

        return type(class_name, class_parents, attr)


class SummarysliceInfo(with_metaclass(Serialisable)):
    """JDAQSummaryslice Metadata.
    """
    h5loc = '/'
    dtype = np.dtype([
        ('det_id', '<i4'),
        ('frame_index', '<u4'),
        ('run_id', '<i4'),
        ('slice_id', '<u4'),
    ])

    @classmethod
    def from_table(cls, row):
        args = []
        for col in cls.dtype.names:
            try:
                args.append(row[col])
            except KeyError:
                args.append(np.nan)
        return cls(*args)

    @classmethod
    def deserialise(cls, *args, **kwargs):
        return cls.conv_from(*args, **kwargs)

    def serialise(self, *args, **kwargs):
        return self.conv_to(*args, **kwargs)

    @classmethod
    def conv_from(cls, data, event_id, fmt='numpy', h5loc='/'):
        if fmt == 'numpy':
            return cls.from_table(data[0])

    def conv_to(self, to='numpy'):
        if to == 'numpy':
            return KM3Array(np.array(self.__array__(), dtype=self.dtype),
                            h5loc=self.h5loc)
        if to == 'pandas':
            return KM3DataFrame(self.conv_to(to='numpy'), h5loc=self.h5loc)

    def __array__(self):
        return [(
            self.det_id, self.frame_index, self.run_id, self.slice_id,
        ), ]

    def __str__(self):
        return "Summaryslice #{0}:\n" \
               "    detector id:     {1}\n" \
               "    frame index:     {2}\n" \
               "    run id:          {3}\n" \
               .format(self.slice_id,
                       self.det_id, self.frame_index, self.run_id,
                       )

    def __insp__(self):
        return self.__str__()

    def __len__(self):
        return 1


class TimesliceInfo(with_metaclass(Serialisable)):
    """JDAQTimeslice metadata.
    """
    h5loc = '/'
    dtype = np.dtype([
        ('dom_id', '<u4'),
        ('frame_id', '<u4'),
        ('n_hits', '<u4'),
        ('slice_id', '<u4'),
    ])

    @classmethod
    def from_table(cls, row):
        args = []
        for col in cls.dtype.names:
            try:
                args.append(row[col])
            except KeyError:
                args.append(np.nan)
        return cls(*args)

    @classmethod
    def deserialise(cls, *args, **kwargs):
        return cls.conv_from(*args, **kwargs)

    def serialise(self, *args, **kwargs):
        return self.conv_to(*args, **kwargs)

    @classmethod
    def conv_from(cls, data, frame_id, fmt='numpy', h5loc='/'):
        if fmt == 'numpy':
            return cls.from_table(data[0])

    def conv_to(self, to='numpy'):
        if to == 'numpy':
            return KM3Array(np.array(self.__array__(), dtype=self.dtype),
                            h5loc=self.h5loc)
        if to == 'pandas':
            return KM3DataFrame(self.conv_to(to='numpy'), h5loc=self.h5loc)

    def __array__(self):
        return [(
            self.dom_id, self.frame_id, self.n_hits, self.slice_id,
        ), ]

    def __str__(self):
        return "Timeslice frame:\n" \
               "    slice id: {0}\n" \
               "    frame id: {1}\n" \
               "    DOM id:   {2}\n" \
               "    n_hits:   {3}\n" \
               .format(self.slice_id, self.frame_id, self.dom_id, self.n_hits)

    def __insp__(self):
        return self.__str__()

    def __len__(self):
        return 1


class TimesliceFrameInfo(with_metaclass(Serialisable)):
    """JDAQTimeslice frame metadata.
    """
    h5loc = '/'
    dtype = np.dtype([
        ('dom_id', '<u4'),
        ('fifo_status', '<u4'),
        ('frame_id', '<u4'),
        ('frame_index', '<u4'),
        ('has_udp_trailer', '<u4'),
        ('high_rate_veto', '<u4'),
        ('max_sequence_number', '<u4'),
        ('n_packets', '<u4'),
        ('slice_id', '<u4'),
        ('utc_nanoseconds', '<u4'),
        ('utc_seconds', '<u4'),
        ('white_rabbit_status', '<u4'),
    ])

    @classmethod
    def from_table(cls, row):
        args = []
        for col in cls.dtype.names:
            try:
                args.append(row[col])
            except KeyError:
                args.append(np.nan)
        return cls(*args)

    @classmethod
    def deserialise(cls, *args, **kwargs):
        return cls.conv_from(*args, **kwargs)

    def serialise(self, *args, **kwargs):
        return self.conv_to(*args, **kwargs)

    @classmethod
    def conv_from(cls, data, frame_id, fmt='numpy', h5loc='/'):
        if fmt == 'numpy':
            return cls.from_table(data[0])

    def conv_to(self, to='numpy'):
        if to == 'numpy':
            return KM3Array(np.array(self.__array__(), dtype=self.dtype),
                            h5loc=self.h5loc)
        if to == 'pandas':
            return KM3DataFrame(self.conv_to(to='numpy'), h5loc=self.h5loc)

    def __array__(self):
        return [(
            self.dom_id, self.fifo_status, self.frame_id, self.frame_index,
            self.has_udp_trailer, self.high_rate_veto,
            self.max_sequence_number, self.n_packets, self.slice_id,
            self.utc_nanoseconds, self.utc_seconds, self.white_rabbit_status
        ), ]

    def __str__(self):
        return "Timeslice frame:\n" \
               "    slice id: {0}\n" \
               "    frame id: {1}\n" \
               "    DOM id:   {2}\n" \
               "    UDP packets: {3}/{4}\n" \
               .format(self.slice_id, self.frame_id, self.dom_id,
                       self.n_packets,
                       self.max_sequence_number)

    def __insp__(self):
        return self.__str__()

    def __len__(self):
        return 1


class SummaryframeInfo(with_metaclass(Serialisable)):
    """JDAQSummaryslice frame metadata.
    """
    h5loc = '/'
    dtype = np.dtype([
        ('dom_id', '<u4'),
        ('fifo_status', '<u4'),
        ('frame_id', '<u4'),
        ('frame_index', '<u4'),
        ('has_udp_trailer', '<u4'),
        ('high_rate_veto', '<u4'),
        ('max_sequence_number', '<u4'),
        ('n_packets', '<u4'),
        ('slice_id', '<u4'),
        ('utc_nanoseconds', '<u4'),
        ('utc_seconds', '<u4'),
        ('white_rabbit_status', '<u4'),
    ])

    @classmethod
    def from_table(cls, row):
        args = []
        for col in cls.dtype.names:
            try:
                args.append(row[col])
            except KeyError:
                args.append(np.nan)
        return cls(*args)

    @classmethod
    def deserialise(cls, *args, **kwargs):
        return cls.conv_from(*args, **kwargs)

    def serialise(self, *args, **kwargs):
        return self.conv_to(*args, **kwargs)

    @classmethod
    def conv_from(cls, data, frame_id, fmt='numpy', h5loc='/'):
        if fmt == 'numpy':
            return cls.from_table(data[0])

    def conv_to(self, to='numpy'):
        if to == 'numpy':
            return KM3Array(np.array(self.__array__(), dtype=self.dtype),
                            h5loc=self.h5loc)
        if to == 'pandas':
            return KM3DataFrame(self.conv_to(to='numpy'), h5loc=self.h5loc)

    def __array__(self):
        return [(
            self.dom_id, self.fifo_status, self.frame_id, self.frame_index,
            self.has_udp_trailer, self.high_rate_veto,
            self.max_sequence_number, self.n_packets, self.slice_id,
            self.utc_nanoseconds, self.utc_seconds, self.white_rabbit_status
        ), ]

    def __str__(self):
        return "Summaryslice frame #{0}:\n" \
               "    slice id:    {1}\n" \
               "    DOM id:      {2}\n" \
               "    UDP packets: {3}/{4}\n" \
               .format(self, self.frame_id, self.slice_id, self.dom_id,
                       self.n_packets,
                       self.max_sequence_number)

    def __insp__(self):
        return self.__str__()

    def __len__(self):
        return 1


class EventInfo(object):
    """Event Metadata.
    """
    dtype = np.dtype([
        ('det_id', '<i4'),
        ('frame_index', '<u4'),
        ('livetime_sec', '<u8'),
        ('mc_id', '<i4'),
        ('mc_t', '<f8'),
        ('n_events_gen', '<u8'),
        ('n_files_gen', '<u8'),
        ('overlays', '<u4'),
        # ('run_id', '<u4'),
        ('trigger_counter', '<u8'),
        ('trigger_mask', '<u8'),
        ('utc_nanoseconds', '<u8'),
        ('utc_seconds', '<u8'),
        ('weight_w1', '<f8'),
        ('weight_w2', '<f8'),
        ('weight_w3', '<f8'),
        ('event_id', '<u4'),
    ])

    def __init__(self, arr, h5loc='/'):
        self._arr = np.array(arr, dtype=self.dtype).reshape(1)
        for col in self.dtype.names:
            setattr(self, col, self._arr[col])
        self.h5loc = h5loc

    @classmethod
    def from_row(cls, row, **kwargs):
        args = tuple((row[col] for col in cls.dtype.names))
        return cls(np.array(args, dtype=cls.dtype), **kwargs)

    @classmethod
    def deserialise(cls, *args, **kwargs):
        return cls.conv_from(*args, **kwargs)

    def serialise(self, *args, **kwargs):
        return self.conv_to(*args, **kwargs)

    @classmethod
    def conv_from(cls, data, event_id, fmt='numpy', h5loc='/', **kwargs):
        if fmt == 'numpy':
            return cls.from_row(data, **kwargs)

    def conv_to(self, to='numpy', **kwargs):
        if to == 'numpy':
            return KM3Array(np.array(self.__array__(), dtype=self.dtype),
                            h5loc=self.h5loc)
        if to == 'pandas':
            return KM3DataFrame(self.conv_to(to='numpy'), h5loc=self.h5loc,
                                **kwargs)

    def __array__(self):
        return self._arr

    def __str__(self):
        return "Event #{0}:\n" \
               "    detector id:     {1}\n" \
               "    frame index:     {2}\n" \
               "    UTC seconds:     {3}\n" \
               "    UTC nanoseconds: {4}\n" \
               "    MC id:           {5}\n" \
               "    MC time:         {6}\n" \
               "    overlays:        {7}\n" \
               "    trigger counter: {8}\n" \
               "    trigger mask:    {9}" \
               .format(self.event_id, self.det_id,
                       self.frame_index, self.utc_seconds,
                       self.utc_nanoseconds, self.mc_id, self.mc_t,
                       self.overlays, self.trigger_counter, self.trigger_mask,
                       # self.run_id,
                       )

    def __repr__(self):
        return self.__str__()

    def __insp__(self):
        return self.__str__()

    def __len__(self):
        return 1


class Point(np.ndarray):
    """Represents a point in a 3D space"""
    def __new__(cls, input_array=(np.nan, np.nan, np.nan)):
        """Add x, y and z to the ndarray"""
        obj = np.asarray(input_array).view(cls)
        return obj

    @property
    def x(self):
        return self[0]

    @x.setter
    def x(self, value):
        self[0] = value

    @property
    def y(self):
        return self[1]

    @y.setter
    def y(self, value):
        self[1] = value

    @property
    def z(self):
        return self[2]

    @z.setter
    def z(self, value):
        self[2] = value


Position = Direction = Point  # Backwards compatibility


class Direction_(Point):
    """Represents a direction in a 3D space

    The direction vector always normalises itself when an attribute is changed.

    """
    def __new__(cls, input_array=(1, 0, 0)):
        """Add x, y and z to the ndarray"""
        normed_array = np.array(input_array) / np.linalg.norm(input_array)
        obj = np.asarray(normed_array).view(cls)
        return obj

    def _normalise(self):
        normed_array = self / np.linalg.norm(self)
        self[0] = normed_array[0]
        self[1] = normed_array[1]
        self[2] = normed_array[2]

    @property
    def x(self):
        return self[0]

    @x.setter
    def x(self, value):
        self[0] = value
        self._normalise()

    @property
    def y(self):
        return self[1]

    @y.setter
    def y(self, value):
        self[1] = value
        self._normalise()

    @property
    def z(self):
        return self[2]

    @z.setter
    def z(self, value):
        self[2] = value
        self._normalise()

    @property
    def zenith(self):
        return angle_between(self, (0, 0, -1))

    def __str__(self):
        return "({0:.4}, {1:.4}, {2:.4})".format(self.x, self.y, self.z)


cdef class Hit:
    """Hit on a PMT.

    Parameters
    ----------
    channel_id : int
    dir_x : float
    dir_y : float
    dir_z : float
    dom_id : int
    id : int
    pmt_id : int
    pos_x : float
    pos_y : float
    pos_z : float
    t0 : int
    time : int
    tot : int
    triggered : int

    """
    cdef public int id, dom_id, time, tot, channel_id, pmt_id
    cdef public unsigned short int triggered
    cdef public float pos_x, pos_y, pos_z, dir_x, dir_y, dir_z, t0

    def __cinit__(self,
                  int channel_id,
                  float dir_x,
                  float dir_y,
                  float dir_z,
                  int dom_id,
                  int id,
                  int pmt_id,
                  float pos_x,
                  float pos_y,
                  float pos_z,
                  int t0,
                  int time,
                  int tot,
                  unsigned short int triggered,
                  int event_id=0        # ignore this!
                  ):
        self.channel_id = channel_id
        self.dir_x = dir_x
        self.dir_y = dir_y
        self.dir_z = dir_z
        self.dom_id = dom_id
        self.id = id
        self.pmt_id = pmt_id
        self.pos_x = pos_x
        self.pos_y = pos_y
        self.pos_z = pos_z
        self.t0 = t0
        self.time = time
        self.tot = tot
        self.triggered = triggered

    @property
    def dir(self):
        return np.array((self.dir_x, self.dir_y, self.dir_z))

    @property
    def pos(self):
        return np.array((self.pos_x, self.pos_y, self.pos_z))

    def __str__(self):
        return "Hit: channel_id({0}), dom_id({1}), pmt_id({2}), tot({3}), " \
               "time({4}), triggered({5})" \
               .format(self.channel_id, self.dom_id, self.pmt_id, self.tot,
                       self.time, self.triggered)

    def __repr__(self):
        return self.__str__()

    def __insp__(self):
        return self.__str__()


cdef class TimesliceHit:
    """Timeslice hit on a PMT.

    Parameters
    ----------
    channel_id : int
    dom_id : int
    time : int
    tot : int

    """
    cdef public int dom_id, time, tot, channel_id
    cdef public np.ndarray pos
    cdef public np.ndarray dir

    def __cinit__(self,
                  int channel_id,
                  int dom_id,
                  int time,
                  int tot,
                  int frame_id=0        # ignore this!
                  ):
        self.channel_id = channel_id
        self.dom_id = dom_id
        self.time = time
        self.tot = tot

    def __str__(self):
        return "TimesliceHit: channel_id({0}), dom_id({1}), tot({2}), " \
               "time({3})" \
               .format(self.channel_id, self.dom_id, self.tot, self.time)

    def __repr__(self):
        return self.__str__()

    def __insp__(self):
        return self.__str__()


cdef class MCHit:
    """Monte Carlo Hit on a PMT.

    Parameters
    ----------
    channel_id : int
    dir : Direction or numpy.ndarray
    dom_id : int
    id : int
    pmt_id : int
    pos : Position or numpy.ndarray
    time : int
    tot : int
    triggered : int

    """
    cdef public int id, dom_id, time, tot, channel_id, pmt_id
    cdef public unsigned short int triggered
    cdef public np.ndarray pos
    cdef public np.ndarray dir

    def __cinit__(self,
                  int channel_id,
                  int dom_id,
                  int id,
                  int pmt_id,
                  int time,
                  int tot,
                  unsigned short int triggered,
                  ):
        self.channel_id = channel_id
        self.dom_id = dom_id
        self.id = id
        self.pmt_id = pmt_id
        self.time = time
        self.tot = tot
        self.triggered = triggered

    def __str__(self):
        return "Hit: channel_id({0}), dom_id({1}), pmt_id({2}), tot({3}), " \
               "time({4}), triggered({5})" \
               .format(self.channel_id, self.dom_id, self.pmt_id, self.tot,
                       self.time, self.triggered)

    def __repr__(self):
        return self.__str__()

    def __insp__(self):
        return self.__str__()


cdef class Track:
    """Particle track.

    Parameters
    ----------
    bjorkeny : float
    dir : Direction or numpy.ndarray
    energy : float
    id : int
    interaction_channel : int
    is_cc : int
    length : float
    pos : Position or numpy.ndarray
    time : int
    type : int
    """
    cdef public int id, time, type, interaction_channel
    cdef public float energy, length, bjorkeny
    cdef public unsigned short int is_cc
    cdef public np.ndarray pos
    cdef public np.ndarray dir

    def __cinit__(self,
                  float bjorkeny,
                  dir,
                  float energy,
                  int id,
                  np.int64_t interaction_channel,
                  unsigned short int is_cc,
                  float length,
                  pos,
                  int time,
                  int type
                  ):
        self.bjorkeny = bjorkeny
        self.is_cc = is_cc
        self.dir = dir
        self.energy = energy
        self.interaction_channel = interaction_channel
        self.id = id
        self.length = length
        self.pos = pos
        self.time = time
        self.type = type

    def __str__(self):
        return "Track: pos({0}), dir({1}), t={2}, E={3}, type={4} ({5})" \
               .format(self.pos, self.dir, self.time, self.energy,
                       self.type, pdg2name(self.type))

    def __repr__(self):
        return self.__str__()

    def __insp__(self):
        return self.__str__()


class HitSeries(object):
    """Collection of multiple Hits.
    """
    dtype = np.dtype([
        ('channel_id', 'u1'),
        ('dir_x', '<f8'),
        ('dir_y', '<f8'),
        ('dir_z', '<f8'),
        ('dom_id', '<u4'),
        ('id', '<u4'),
        ('pmt_id', '<u4'),
        ('pos_x', '<f8'),
        ('pos_y', '<f8'),
        ('pos_z', '<f8'),
        # ('run_id', '<u4'),
        ('t0', '<i4'),
        ('time', '<i4'),
        ('tot', 'u1'),
        ('triggered', 'u1'),
        ('event_id', '<u4'),
    ])

    def __init__(self, arr, h5loc='/'):
        self._arr = arr
        self._index = 0
        self._hits = None
        self.h5loc = h5loc

    @classmethod
    def from_aanet(cls, hits, event_id):
        try:
            return cls(np.array([(
                h.channel_id,
                np.nan,     # h.dir.x,
                np.nan,     # h.dir.y,
                np.nan,     # h.dir.z,
                h.dom_id,
                h.id,
                h.pmt_id,
                np.nan,     # h.pos.x,
                np.nan,     # h.pos.y,
                np.nan,     # h.pos.z,
                0,          # t0
                h.t,
                h.tot,
                h.trig,
                event_id,
            ) for h in hits], dtype=cls.dtype))
        except ValueError:
            # Older aanet format.
            return cls(np.array([(
                ord(h.channel_id),
                np.nan,     # h.dir.x,
                np.nan,     # h.dir.y,
                np.nan,     # h.dir.z,
                h.dom_id,
                h.id,
                h.pmt_id,
                np.nan,     # h.pos.x,
                np.nan,     # h.pos.y,
                np.nan,     # h.pos.z,
                0,          # t0
                h.t,
                h.tot,
                h.trig,
                event_id,
            ) for h in hits], dtype=cls.dtype))

    @classmethod
    def from_evt(cls, hits, event_id):
        return cls(np.array([(
            0,     # channel_id
            np.nan,
            np.nan,
            np.nan,
            0,     # dom_id
            h.id,
            h.pmt_id,
            np.nan,
            np.nan,
            np.nan,
            0,      # t0
            h.time,
            h.tot,
            0,     # triggered
            event_id,
        ) for h in hits], dtype=cls.dtype))

    @classmethod
    def from_arrays(cls, channel_ids, dir_xs, dir_ys, dir_zs, dom_ids, ids,
                    pmt_ids, pos_xs, pos_ys, pos_zs, t0s, times, tots,
                    triggereds, event_id):
        # do we need shape[0] or does len() work too?
        try:
            length = channel_ids.shape[0]
        except AttributeError:
            length = len(channel_ids)
        hits = np.empty(length, cls.dtype)
        hits['channel_id'] = channel_ids
        hits['dir_x'] = dir_xs
        hits['dir_y'] = dir_ys
        hits['dir_z'] = dir_zs
        hits['dom_id'] = dom_ids
        hits['id'] = ids
        hits['pmt_id'] = pmt_ids
        hits['pos_x'] = pos_xs
        hits['pos_y'] = pos_ys
        hits['pos_z'] = pos_zs
        hits['t0'] = t0s
        hits['time'] = times
        hits['tot'] = tots
        hits['triggered'] = triggereds
        hits['event_id'] = np.full(length, event_id, dtype='<u4')
        return cls(hits)

    @classmethod
    def from_dict(cls, map, event_id):
        if event_id is None:
            event_id = map['event_id']
        return cls.from_arrays(
            map['channel_id'],
            map['dir_x'],
            map['dir_y'],
            map['dir_z'],
            map['dom_id'],
            map['id'],
            map['pmt_id'],
            map['pos_x'],
            map['pos_y'],
            map['pos_z'],
            map['t0'],
            map['time'],
            map['tot'],
            map['triggered'],
            event_id,
        )

    @classmethod
    def from_table(cls, table, event_id):
        if event_id is None:
            event_id = table[0]['event_id']
        return cls(np.array([(
            row['channel_id'],
            row['dir_x'],
            row['dir_y'],
            row['dir_z'],
            row['dom_id'],
            row['id'],
            row['pmt_id'],
            row['pos_x'],
            row['pos_y'],
            row['pos_z'],
            row['t0'],
            row['time'],
            row['tot'],
            row['triggered'],
        ) for row in table], dtype=cls.dtype), event_id)

    @classmethod
    def deserialise(cls, *args, **kwargs):
        return cls.conv_from(*args, **kwargs)

    def serialise(self, *args, **kwargs):
        return self.conv_to(*args, **kwargs)

    @classmethod
    def conv_from(cls, data, event_id=None, fmt='numpy', h5loc='/'):
        # what is event_id doing here?
        if fmt == 'numpy':
            return cls(data)
        if fmt == 'pandas':
            return cls(data.to_records(index=False))

    def conv_to(self, to='numpy'):
        if to == 'numpy':
            return KM3Array(np.array(self.__array__(), dtype=self.dtype),
                            h5loc=self.h5loc)
        if to == 'pandas':
            return KM3DataFrame(self.conv_to(to='numpy'), h5loc=self.h5loc)

    def __array__(self):
        return self._arr

    def __iter__(self):
        return self

    @property
    def id(self):
        return self._arr['id']

    @property
    def time(self):
        return self._arr['time']

    @property
    def triggered_hits(self):
        return HitSeries(self._arr[self._arr['triggered'] == True])     # noqa

    @property
    def first_hits(self):
        h = self.conv_to(to='pandas')
        h.sort_values('time', inplace=True)
        h = h.drop_duplicates(subset='dom_id')
        return HitSeries(h.to_records(index=False))

    @property
    def triggered(self):
        return self._arr['triggered']

    @property
    def tot(self):
        return self._arr['tot']

    @property
    def dom_id(self):
        return self._arr['dom_id']

    @property
    def pmt_id(self):
        return self._arr['pmt_id']

    @property
    def id(self):
        return self._arr['id']

    @property
    def channel_id(self):
        return self._arr['channel_id']

    @property
    def pos_x(self):
        return self._arr['pos_x']

    @property
    def pos_y(self):
        return self._arr['pos_y']

    @property
    def pos_z(self):
        return self._arr['pos_z']

    @property
    def dir_x(self):
        return self._arr['dir_x']

    @property
    def dir_y(self):
        return self._arr['dir_y']

    @property
    def dir_z(self):
        return self._arr['dir_z']

    @property
    def t0(self):
        return self._arr['t0']

    def next(self):
        """Python 2/3 compatibility for iterators"""
        return self.__next__()

    def __next__(self):
        if self._index >= len(self):
            self._index = 0
            raise StopIteration
        hit = self[self._index]
        self._index += 1
        return hit

    def __len__(self):
        return self._arr.shape[0]

    def __getitem__(self, index):
        if isinstance(index, int):
            hit = Hit(*self._arr[index])
            return hit
        elif isinstance(index, slice):
            return self._slice_generator(index)
        else:
            raise TypeError("index must be int or slice")

    def _slice_generator(self, index):
        """A simple slice generator for iterations"""
        start, stop, step = index.indices(len(self))
        for i in range(start, stop, step):
            yield Hit(*self._arr[i])

    def __str__(self):
        n_hits = len(self)
        plural = 's' if n_hits > 1 or n_hits == 0 else ''
        return("HitSeries with {0} hit{1}.".format(len(self), plural))

    def __repr__(self):
        return self.__str__()

    def __insp__(self):
        return '\n'.join([str(hit) for hit in self._hits])


class TimesliceHitSeries(object):
    """Collection of multiple timeslice hits.
    """
    dtype = np.dtype([
        ('channel_id', 'u1'),
        ('dom_id', '<u4'),
        ('time', '<i4'),
        ('tot', 'u1'),
    ])

    def __init__(self, arr, slice_id, frame_id):
        self._arr = arr
        self._index = 0
        self._hits = None
        self.slice_id = slice_id
        self.frame_id = frame_id

    @property
    def h5loc(self):
        return "/timeslices/slice_{0}".format(self.slice_id)

    @property
    def tabname(self):
        return "frame_{0}".format(self.frame_id)

    @classmethod
    def from_arrays(cls, channel_ids, dom_ids, times, tots, slice_id,
                    frame_id):
        len = channel_ids.shape[0]
        hits = np.empty(len, cls.dtype)
        hits['channel_id'] = channel_ids
        hits['dom_id'] = dom_ids
        hits['time'] = times
        hits['tot'] = tots
        return cls(hits, slice_id, frame_id)

    @classmethod
    def from_dict(cls, map, slice_id, frame_id):
        return cls.from_arrays(
            map['channel_id'],
            map['dom_id'],
            map['time'],
            map['tot'],
            slice_id,
            frame_id,
        )

    def from_table(cls, table, slice_id, frame_id):
        return cls(np.array([(
            row['channel_id'],
            row['dom_id'],
            row['time'],
            row['tot'],
        ) for row in table], dtype=cls.dtype), slice_id, frame_id)

    @classmethod
    def deserialise(cls, *args, **kwargs):
        return cls.conv_from(*args, **kwargs)

    def serialise(self, *args, **kwargs):
        return self.conv_to(*args, **kwargs)

    @classmethod
    def conv_from(cls, data, slice_id, frame_id, fmt='numpy', h5loc='/'):
        if fmt == 'numpy':
            # return cls.from_table(data, frame_id)
            return cls(data, slice_id, frame_id)

    def conv_to(self, to='numpy'):
        if to == 'numpy':
            return KM3Array(np.array(self.__array__(), dtype=self.dtype),
                            h5loc=self.h5loc)
        if to == 'pandas':
            return KM3DataFrame(self.conv_to(to='numpy'), h5loc=self.h5loc)

    def __array__(self):
        return self._arr

    def __iter__(self):
        return self

    @property
    def time(self):
        return self._arr['time']

    @property
    def tot(self):
        return self._arr['tot']

    @property
    def dom_id(self):
        return self._arr['dom_id']

    @property
    def channel_id(self):
        return self._arr['channel_id']

    def next(self):
        """Python 2/3 compatibility for iterators"""
        return self.__next__()

    def __next__(self):
        if self._index >= len(self):
            self._index = 0
            raise StopIteration
        hits = TimesliceHit(*self._arr[self._index])
        self._index += 1
        return hits

    def __len__(self):
        return self._arr.shape[0]

    def __getitem__(self, index):
        if isinstance(index, int):
            return TimesliceHit(*self._arr[index])
        elif isinstance(index, slice):
            return self._slice_generator(index)
        else:
            raise TypeError("index must be int or slice")

    def _slice_generator(self, index):
        """A simple slice generator for iterations"""
        start, stop, step = index.indices(len(self))
        for i in range(start, stop, step):
            yield TimesliceHit(*self._arr[i])

    def __str__(self):
        n_hits = len(self)
        plural = 's' if n_hits > 1 or n_hits == 0 else ''
        return("TimesliceHitSeries with {0} hit{1}.".format(len(self), plural))

    def __repr__(self):
        return self.__str__()

    def __insp__(self):
        return '\n'.join([str(hit) for hit in self._hits])


class TrackSeries(object):
    """Collection of multiple Tracks.

    Attributes
    ----------
    dtype: datatype of array representation
    """
    dtype = np.dtype([
        ('bjorkeny', '<f8'),
        ('dir_x', '<f8'),
        ('dir_y', '<f8'),
        ('dir_z', '<f8'),
        ('energy', '<f8'),
        ('id', '<u4'),
        ('interaction_channel', '<u4'),
        ('is_cc', '<u4'),
        ('length', '<f8'),
        ('pos_x', '<f8'),
        ('pos_y', '<f8'),
        ('pos_z', '<f8'),
        # ('run_id', '<u4'),
        ('time', '<i4'),
        ('type', '<i4'),
        ('event_id', '<u4'),
    ])

    def __init__(self, tracks, event_id, h5loc='/'):
        self._bjorkeny = None
        self._dir = None
        self._energy = None
        self._highest_energetic_muon = None
        self._id = None
        self._index = 0
        self._interaction_channel = None
        self._is_cc = None
        self._length = None
        self._pos = None
        self._time = None
        self._tracks = tracks
        self._type = None
        self.event_id = event_id
        self.h5loc = h5loc

    @classmethod
    def from_aanet(cls, tracks, event_id):
        return cls([Track(cls.get_usr_item(t, 1),               # bjorkeny
                          Direction((t.dir.x, t.dir.y, t.dir.z)),
                          t.E,
                          t.id,
                          cls.get_usr_item(t, 2),               # ichan
                          IS_CC[cls.get_usr_item(t, 0)],        # is_cc
                          cls.get_len(t),
                          Position((t.pos.x, t.pos.y, t.pos.z)),
                          t.t,
                          # This is a nasty bug. It is not completely clear
                          # if this is supposed to be PDG or Geant convention.
                          # might be, that for CC neutrino events,
                          # the two vector elements might follow _different_
                          # conventions. Yep, 2 conventions for
                          # 2 vector elements...
                          #
                          # geant2pdg(t.type))
                          t.type,
                          )
                    for t in tracks], event_id)

    @classmethod
    def from_arrays(cls,
                    bjorkenys,
                    directions_x,
                    directions_y,
                    directions_z,
                    energies,
                    ids,
                    interaction_channels,
                    is_ccs,
                    lengths,
                    positions_x,
                    positions_y,
                    positions_z,
                    times,
                    types,
                    event_id,
                    ):
        directions = np.column_stack((directions_x, directions_y,
                                      directions_z))
        positions = np.column_stack((positions_x, positions_y,
                                     positions_z))
        args = bjorkenys, directions, energies, \
            ids, interaction_channels, is_ccs, lengths, positions, \
            times, types
        tracks = cls([Track(*track_args) for track_args in zip(*args)],
                     event_id)
        tracks._bjorkeny = bjorkenys
        tracks._dir = zip(directions_x, directions_y, directions_z)
        tracks._energy = energies
        tracks._id = ids
        tracks._interaction_channel = interaction_channels
        tracks._is_cc = is_ccs
        tracks._length = lengths
        tracks._pos = zip(positions_x, positions_y, positions_z)
        tracks._time = times
        tracks._type = types
        return tracks

    @classmethod
    def from_km3df(cls, km3df):
        return cls.from_table(km3df.conv_to(to='numpy'),
                              event_id=int(km3df['event_id']))

    @classmethod
    def from_table(cls, table, event_id):
        return cls([Track(
            row['bjorkeny'],
            np.array((row['dir_x'], row['dir_y'], row['dir_z'],)),
            row['energy'],
            row['id'],
            row['interaction_channel'],
            row['is_cc'],
            row['length'],
            np.array((row['pos_x'], row['pos_y'], row['pos_z'],)),
            row['time'],
            row['type']
        ) for row in table], event_id)

    @classmethod
    def deserialise(cls, *args, **kwargs):
        return cls.conv_from(*args, **kwargs)

    def serialise(self, *args, **kwargs):
        return self.conv_to(*args, **kwargs)

    @classmethod
    def conv_from(cls, data, event_id, fmt='numpy', h5loc='/'):
        if fmt == 'numpy':
            return cls.from_table(data, event_id)

    def conv_to(self, to='numpy'):
        if to == 'numpy':
            return KM3Array(np.array(self.__array__(), dtype=self.dtype),
                            h5loc=self.h5loc)
        if to == 'pandas':
            return KM3DataFrame(self.conv_to(to='numpy'), h5loc=self.h5loc)

    def __array__(self):
        return [(
            t.bjorkeny, t.dir[0], t.dir[1], t.dir[2], t.energy,
            t.id, t.interaction_channel, t.is_cc,
            t.length, t.pos[0], t.pos[1], t.pos[2], t.time, t.type,
            self.event_id,
        ) for t in self._tracks]

    @classmethod
    def get_len(cls, track):
        try:
            return track.len
        except AttributeError:
            return 0

    @classmethod
    def get_usr_item(cls, track, index):
        try:
            item = track.usr[index]
        except IndexError:
            item = 0.
        return item

    @property
    def highest_energetic_muon(self):
        if self._highest_energetic_muon is None:
            muons = [track for track in self if abs(track.type) == 13]
            if len(muons) == 0:
                raise AttributeError("No muon found")
            self._highest_energetic_muon = max(muons, key=lambda m: m.energy)
        return self._highest_energetic_muon

    def __iter__(self):
        return self

    def __iter__(self):
        return self

    @property
    def bjorkeny(self):
        if self._bjorkeny is None:
            self._bjorkeny = np.array([t.bjorkeny for t in self._tracks])
        return self._bjorkeny

    @property
    def is_cc(self):
        if self._is_cc is None:
            self._is_cc = np.array([t.is_cc for t in self._tracks])
        return self._is_cc

    @property
    def interaction_channel(self):
        if self._interaction_channel is None:
            self._interaction_channel = np.array([t.interaction_channel for
                                                  t in self._tracks])
        return self._interaction_channel

    @property
    def id(self):
        if self._id is None:
            self._id = np.array([t.id for t in self._tracks])
        return self._id

    @property
    def time(self):
        if self._time is None:
            self._time = np.array([t.time for t in self._tracks])
        return self._time

    @property
    def energy(self):
        if self._energy is None:
            self._energy = np.array([t.energy for t in self._tracks])
        return self._energy

    @property
    def length(self):
        if self._length is None:
            self._length = np.array([t.length for t in self._tracks])
        return self._length

    @property
    def type(self):
        if self._type is None:
            self._type = np.array([t.type for t in self._tracks])
        return self._type

    @property
    def pos(self):
        if self._pos is None:
            self._pos = np.array([t.pos for t in self._tracks])
        return self._pos

    @property
    def dir(self):
        if self._dir is None:
            self._dir = np.array([t.dir for t in self._tracks])
        return self._dir

    def next(self):
        """Python 2/3 compatibility for iterators"""
        return self.__next__()

    def __next__(self):
        if self._index >= len(self):
            self._index = 0
            raise StopIteration
        item = self._tracks[self._index]
        self._index += 1
        return item

    def __len__(self):
        return len(self._tracks)

    def __getitem__(self, index):
        if isinstance(index, int):
            return self._tracks[index]
        elif isinstance(index, slice):
            return self._slice_generator(index)
        else:
            raise TypeError("index must be int or slice")

    def _slice_generator(self, index):
        """A simple slice generator for iterations"""
        start, stop, step = index.indices(len(self))
        for i in range(start, stop, step):
            yield self._tracks[i]

    def __str__(self):
        n_tracks = len(self)
        plural = 's' if n_tracks > 1 or n_tracks == 0 else ''
        return("TrackSeries with {0} track{1}.".format(len(self), plural))

    def __repr__(self):
        return self.__str__()

    def __insp__(self):
        return '\n'.join([str(track) for track in self._tracks])


class SummaryframeSeries(object):
    """Collection of summary frames.
    """
    dtype = np.dtype([
        ('dom_id', '<u4'),
        ('max_sequence_number', '<u4'),
        ('n_received_packets', '<u4'),
        ('slice_id', '<u4'),
        ])
    h5loc = '/'

    def __init__(self, arr):
        self._arr = arr
        self._index = 0
        self._frames = None

    @classmethod
    def from_arrays(cls, dom_ids, max_sequence_numbers, n_received_packets,
                    slice_id):
        length = dom_ids.shape[0]
        frames = np.empty(length, cls.dtype)
        frames['dom_id'] = dom_ids
        frames['max_sequence_number'] = max_sequence_numbers
        frames['n_received_packets'] = n_received_packets
        frames['slice_id'] = np.full(length, slice_id, dtype='<u4')
        return cls(frames)

    def from_table(cls, table, slice_id):
        if slice_id is None:
            slice_id = table[0]['slice_id']
        return cls(np.array([(
            row['dom_id'],
            row['max_sequence_number'],
            row['n_received_packets'],
        ) for row in table], dtype=cls.dtype), slice_id)

    @classmethod
    def deserialise(cls, *args, **kwargs):
        return cls.conv_from(*args, **kwargs)

    def serialise(self, *args, **kwargs):
        return self.conv_to(*args, **kwargs)

    @classmethod
    def conv_from(cls, data, slice_id, fmt='numpy', h5loc='/'):
        if fmt == 'numpy':
            # return cls.from_table(data, event_id)
            return cls(data)

    def conv_to(self, to='numpy'):
        if to == 'numpy':
            return KM3Array(np.array(self.__array__(), dtype=self.dtype),
                            h5loc=self.h5loc)
        if to == 'pandas':
            return KM3DataFrame(self.conv_to(to='numpy'), h5loc=self.h5loc)

    @property
    def n_received_packets(self):
        return self._arr['n_received_packets']

    @property
    def dom_ids(self):
        return self._arr['dom_ids']

    @property
    def max_sequence_numbers(self):
        return self._arr['max_sequence_numbers']

    def __array__(self):
        return self._arr

    def __iter__(self):
        return self

    def next(self):
        """Python 2/3 compatibility for iterators"""
        return self.__next__()

    def __next__(self):
        if self._index >= len(self):
            self._index = 0
            raise StopIteration
        data = self._arr[self._index]
        self._index += 1
        return data

    def __len__(self):
        return self._arr.shape[0]

    def __getitem__(self, index):
        if isinstance(index, int):
            return self._arr[index]
        elif isinstance(index, slice):
            return self._slice_generator(index)
        else:
            raise TypeError("index must be int or slice")

    def _slice_generator(self, index):
        """A simple slice generator for iterations"""
        start, stop, step = index.indices(len(self))
        for i in range(start, stop, step):
            yield self._arr[i]


class KM3Array(np.ndarray):
    """Numpy NDarray + metadata.

    This class adds the following to ``np.ndarray``:

    Attributes
    ----------
    h5loc: str, default='/'
        HDF5 group where to write into.

    Methods
    -------
    deserialise(data, h5loc='/', fmt='pandas')
        Factory to init from data.
    serialise(to='numpy')
        Convert for storage.
    """
    def __new__(cls, arr, h5loc='/'):
        obj = np.asarray(arr).view(cls)
        obj.h5loc = h5loc
        return obj

    def __init__(self, arr, h5loc='/'):
        self.array = arr
        self.h5loc = h5loc

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.h5loc = getattr(obj, 'h5loc', None)

    @classmethod
    def deserialise(cls, *args, **kwargs):
        return cls.conv_from(*args, **kwargs)

    def serialise(self, *args, **kwargs):
        return self.conv_to(*args, **kwargs)

    @classmethod
    def conv_from(cls, data, event_id=None, h5loc='/', fmt='numpy', **kwargs):
        if fmt == 'numpy':
            arr = cls(data, h5loc, **kwargs)
            return arr

    def conv_to(self, to='numpy', **kwargs):
        if to == 'numpy':
            return self
        if to == 'pandas':
            return KM3DataFrame.conv_from(self, fmt='numpy', **kwargs)

    @classmethod
    def from_dict(cls, map, dtype=None, **kwargs):
        if dtype is None:
            dtype = np.dtype([(key, float) for key in sorted(map.keys())])
        return cls(np.array([tuple((map[key] for key in dtype.names))],
                            dtype=dtype), **kwargs)


class KM3DataFrame(pd.DataFrame):
    """Pandas Dataframe + metadata.

    This class adds the following to ``pd.DataFrame``:

    Attributes
    ----------
    h5loc: str, default='/'
        HDF5 group where to write into.

    Methods
    -------
    deserialise(data, h5loc='/', fmt='pandas')
        Factory to init from data.
    serialise(to='numpy')
        Convert for storage.
    """

    # do not rename this!
    # http://pandas.pydata.org/pandas-docs/stable/internals.html#define-original-properties
    # this is preserved over df manipulations
    _metadata = ['h5loc']

    # default value
    h5loc = '/'

    @property
    def _constructor(self):
        return KM3DataFrame

    def __init__(self, *args, **kwargs):
        h5loc = kwargs.pop('h5loc', '/')
        super(KM3DataFrame, self).__init__(*args, **kwargs)
        self.h5loc = h5loc

    @classmethod
    def deserialise(cls, *args, **kwargs):
        return cls.conv_from(*args, **kwargs)

    def serialise(self, *args, **kwargs):
        return self.conv_to(*args, **kwargs)

    @classmethod
    def conv_from(cls, data, event_id=None, h5loc='/', fmt='pandas', **kwargs):
        if fmt in {'numpy', 'pandas'}:
            return cls(data, h5loc=h5loc, **kwargs)
        if fmt == 'dict':
            return cls(data, index=[0], h5loc=h5loc, **kwargs)

    def conv_to(self, to='numpy', **kwargs):
        if to == 'numpy':
            return KM3Array(self.to_records(index=False), h5loc=self.h5loc,
                            **kwargs)
        if to == 'pandas':
            return self


deserialise_map = {
    'McHits': HitSeries,
    'Hits': HitSeries,
    'TimesliceHits': TimesliceHitSeries,
    'McTracks': TrackSeries,
    'EventInfo': EventInfo,
    'SummarysliceInfo': SummarysliceInfo,
    'SummaryframeInfo': SummaryframeInfo,
    'Tracks': TrackSeries,
}


class BinaryStruct(object):
    """A binary struct superclass which parses itself from a given stream.

    Overwrite `_structure` and `_fields` after subclassing.
    `_structure` should have the standard Python struct format type syntax.
    The `_fields` are automatically translated into instance attributes
    at init.

    Parameters
    ----------
    _structure: struct-format as str, like '<i2f'
    _fields: iterable of str

    """
    _structure = ''
    _fields = ()

    def __init__(self, stream):
        data = Struct(self._structure).unpack_from(stream.read(self._size))
        for attr, value in zip(self._fields, data):
            setattr(self, attr, value)

    @property
    def _size(self):
        return calcsize(self._structure)



# coding=utf-8
# Filename: fit.py
# pylint: disable=locally-disabled
"""
A collection of fit functions and modules.

"""
from __future__ import division, absolute_import, print_function

from scipy import optimize
import numpy as np


def fit_delta_ts(data, time_s):
    """Fits gaussians to delta t for each PMT pair.

    Parameters
    ----------
    data: 2d np.array: x = PMT combinations (465), y = time, entry = frequency
    time_s: length of data taking in seconds

    Returns
    -------
    numpy arrays with rates and means for all PMT combinations
    """

    data = data / time_s

    xs = np.arange(-20, 21)
    def gaussian(x, mean, sigma, rate, offset):
        return rate / np.sqrt(2 * np.pi) /  \
               sigma * np.exp(-(x-mean)**2 / sigma**2) + offset
    rates = []
    means = []
    for combination in data:
        try:
            popt, _ = optimize.curve_fit(gaussian, xs, combination,
                                         p0=[0, 2, 1000, 20])
        except RuntimeError:
            popt = (0, 0, 0, 0)
        rates.append(popt[2])
        means.append(popt[0])
    return np.array(rates), np.array(means)



# coding=utf-8
# Filename: h5tree.py
"""
Print the HDF5 file structure.

Usage:
    h5tree FILE
    h5tree (-h | --help)
    h5tree --version

Options:
    FILE       Input file.
    -h --help  Show this screen.

"""
from __future__ import division, absolute_import, print_function

import tables

from km3pipe.tools import deprecated

__author__ = "Moritz Lotze"
__copyright__ = "Copyright 2016, Moritz Lotze and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Moritz Lotze"
__email__ = "mlotze@km3net.de"
__status__ = "Development"


MSG = "Use `ptdump` from pytables instead, it's better!"

@deprecated(MSG)
def h5tree(h5name):
    with tables.open_file(h5name) as h5:
        for node in h5.walk_nodes():
            print(node)


def main():
    from docopt import docopt
    arguments = docopt(__doc__)

    print('DeprecationWarning: h5tree.py is deprecated. ' + MSG)
    h5tree(arguments['FILE'])



# coding=utf-8
# Filename: test_hardware.py
# pylint: disable=C0111,C0103,R0904
"""
Detector description (detx format v5)

global_det_id ndoms
dom_id line_id floor_id npmts
 pmt_id_global x y z dx dy dz t0
 pmt_id_global x y z dx dy dz t0
 ...
 pmt_id_global x y z dx dy dz t0
dom_id line_id floor_id npmts
 ...

"""
from __future__ import division, absolute_import, print_function

from km3pipe.testing import TestCase, StringIO, skipIf
from km3pipe.hardware import Detector, PMT

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


EXAMPLE_DETX = StringIO("\n".join((
    "1 3",
    "1 1 1 3",
    " 1 1.1 1.2 1.3 -1.1  0.2  0.3 10",
    " 2 1.4 1.5 1.6  0.1 -1.2  0.3 20",
    " 3 1.7 1.8 1.9  0.1  0.2 -1.3 30",
    "2 1 2 3",
    " 4 2.1 2.2 2.3 -1.1  0.2  0.3 40",
    " 5 2.4 2.5 2.6  0.1 -1.2  0.3 50",
    " 6 2.7 2.8 2.9  0.1  0.2 -1.3 60",
    "3 1 3 3",
    " 7 3.1 3.2 3.3 -1.1  0.2  0.3 70",
    " 8 3.4 3.5 3.6  0.1 -1.2  0.3 80",
    " 9 3.7 3.8 3.9  0.1  0.2 -1.3 90",)))

EXAMPLE_DETX_MIXED_IDS = StringIO("\n".join((
    "1 3",
    "8 1 1 3",
    " 83 1.1 1.2 1.3 -1.1  0.2  0.3 10",
    " 81 1.4 1.5 1.6  0.1 -1.2  0.3 20",
    " 82 1.7 1.8 1.9  0.1  0.2 -1.3 30",
    "7 1 2 3",
    " 71 2.1 2.2 2.3 -1.1  0.2  0.3 40",
    " 73 2.4 2.5 2.6  0.1 -1.2  0.3 50",
    " 72 2.7 2.8 2.9  0.1  0.2 -1.3 60",
    "6 1 3 3",
    " 62 3.1 3.2 3.3 -1.1  0.2  0.3 70",
    " 63 3.4 3.5 3.6  0.1 -1.2  0.3 80",
    " 61 3.7 3.8 3.9  0.1  0.2 -1.3 90",)))

EXAMPLE_DETX_WRITE = StringIO("\n".join((
    "1 3",
    "1 1 1 3",
    " 1 1.1 1.2 1.3 1.0 0.0 0.0 10.0",
    " 2 1.4 1.5 1.6 0.0 1.0 0.0 20.0",
    " 3 1.7 1.8 1.9 0.0 0.0 1.0 30.0",
    "2 1 2 3",
    " 4 2.1 2.2 2.3 0.0 1.0 0.0 40.0",
    " 5 2.4 2.5 2.6 0.0 0.0 1.0 50.0",
    " 6 2.7 2.8 2.9 1.0 0.0 0.0 60.0",
    "3 1 3 3",
    " 7 3.1 3.2 3.3 0.0 0.0 1.0 70.0",
    " 8 3.4 3.5 3.6 0.0 1.0 0.0 80.0",
    " 9 3.7 3.8 3.9 1.0 0.0 0.0 90.0\n",)))

EXAMPLE_MC_DETX_WRITE_MIXED_IDS = StringIO("\n".join((
    "-1 3",
    "6 1 1 3",
    " 31 1.1 1.2 1.3 1.0 0.0 0.0 10.0",
    " 22 1.4 1.5 1.6 0.0 1.0 0.0 20.0",
    " 13 1.7 1.8 1.9 0.0 0.0 1.0 30.0",
    "3 1 2 3",
    " 34 2.1 2.2 2.3 1.0 0.0 0.0 40.0",
    " 45 2.4 2.5 2.6 0.0 1.0 0.0 50.0",
    " 16 2.7 2.8 2.9 0.0 0.0 1.0 60.0",
    "9 1 3 3",
    " 17 3.1 3.2 3.3 1.0 0.0 0.0 70.0",
    " 48 3.4 3.5 3.6 0.0 1.0 0.0 80.0",
    " 39 3.7 3.8 3.9 0.0 0.0 1.0 90.0\n",)))


class TestDetector(TestCase):

    def setUp(self):
        self.det = Detector()
        self.det._det_file = EXAMPLE_DETX

    def test_parse_header_extracts_correct_det_id(self):
        self.det._parse_header()
        self.assertEqual(1, self.det.det_id)

    def test_parse_header_extracts_correct_n_doms(self):
        self.det._parse_header()
        self.assertEqual(3, self.det.n_doms)

    def test_parse_doms_maps_each_dom_correctly(self):
        self.det._parse_doms()
        expected = {1: (1, 1, 3), 2: (1, 2, 3), 3: (1, 3, 3)}
        self.assertDictEqual(expected, self.det.doms)

    def test_dom_ids(self):
        self.det._parse_doms()
        self.assertEqual((1, 2, 3), tuple(self.det.dom_ids))

    def test_parse_doms_maps_each_dom_correctly_for_mixed_pmt_ids(self):
        self.det._det_file = EXAMPLE_DETX_MIXED_IDS
        self.det._parse_doms()
        expected = {8: (1, 1, 3), 7: (1, 2, 3), 6: (1, 3, 3)}
        self.assertDictEqual(expected, self.det.doms)

    def test_dom_positions(self):
        self.det._parse_doms()
        self.assertAlmostEqual(1.4, self.det.dom_positions[1][0])
        self.assertAlmostEqual(1.5, self.det.dom_positions[1][1])
        self.assertAlmostEqual(1.6, self.det.dom_positions[1][2])
        self.assertAlmostEqual(2.4, self.det.dom_positions[2][0])
        self.assertAlmostEqual(2.5, self.det.dom_positions[2][1])
        self.assertAlmostEqual(2.6, self.det.dom_positions[2][2])

    def test_pmt_with_id_returns_correct_omkeys(self):
        self.det._parse_doms()
        self.assertEqual((1, 1, 0), self.det.pmt_with_id(1).omkey)
        self.assertEqual((1, 2, 1), self.det.pmt_with_id(5).omkey)

    def test_pmt_with_id_returns_correct_omkeys_with_mixed_pmt_ids(self):
        self.det._det_file = EXAMPLE_DETX_MIXED_IDS
        self.det._parse_doms()
        self.assertEqual((1, 2, 1), self.det.pmt_with_id(73).omkey)
        self.assertEqual((1, 1, 1), self.det.pmt_with_id(81).omkey)

    def test_pmt_with_id_raises_exception_for_invalid_id(self):
        self.det._parse_doms()
        with self.assertRaises(KeyError):
            self.det.pmt_with_id(100)

    def test_get_pmt(self):
        self.det._det_file = EXAMPLE_DETX_MIXED_IDS
        self.det._parse_doms()
        pmt = self.det.get_pmt(7, 2)
        self.assertEqual((1, 2, 2), pmt.omkey)

    def test_pmtid2omkey_old(self):
        pmtid2omkey = self.det._pmtid2omkey_old
        self.assertEqual((1, 13, 12), tuple(pmtid2omkey(168)))
        self.assertEqual((1, 12, 18), tuple(pmtid2omkey(205)))
        self.assertEqual((1, 11, 22), tuple(pmtid2omkey(240)))
        self.assertEqual((4, 11, 2), tuple(pmtid2omkey(1894)))
        self.assertEqual((9, 18, 0), tuple(pmtid2omkey(4465)))
        self.assertEqual((95, 7, 16), tuple(pmtid2omkey(52810)))
        self.assertEqual((95, 4, 13), tuple(pmtid2omkey(52900)))

    def test_pmtid2omkey_old_handles_floats(self):
        pmtid2omkey = self.det._pmtid2omkey_old
        self.assertEqual((1, 13, 12), tuple(pmtid2omkey(168.0)))
        self.assertEqual((1, 12, 18), tuple(pmtid2omkey(205.0)))
        self.assertEqual((1, 11, 22), tuple(pmtid2omkey(240.0)))
        self.assertEqual((4, 11, 2), tuple(pmtid2omkey(1894.0)))
        self.assertEqual((9, 18, 0), tuple(pmtid2omkey(4465.0)))
        self.assertEqual((95, 7, 16), tuple(pmtid2omkey(52810.0)))
        self.assertEqual((95, 4, 13), tuple(pmtid2omkey(52900.0)))


class TestPMT(TestCase):

    def test_init(self):
        pmt = PMT(1, (1., 2, 3), (4., 5, 6), 7, 8, (9, 10, 11))
        self.assertAlmostEqual(1, pmt.id)
        self.assertAlmostEqual(1, pmt.pos.x)
        self.assertAlmostEqual(2, pmt.pos.y)
        self.assertAlmostEqual(3, pmt.pos.z)
        self.assertAlmostEqual(4, pmt.dir.x, 6)
        self.assertAlmostEqual(5, pmt.dir.y, 6)
        self.assertAlmostEqual(6, pmt.dir.z, 6)
        self.assertAlmostEqual(7, pmt.t0)
        self.assertAlmostEqual(8, pmt.channel_id)
        self.assertEqual((9, 10, 11), pmt.omkey)



# coding=utf-8
# Filename: astro.py
# pylint: disable=C0103
# pragma: no cover
"""
Astro utils.
"""
from __future__ import division, absolute_import, print_function

import numpy as np
from astropy.units import degree, rad
from astropy.coordinates import (Latitude, Longitude,
                                 EarthLocation, AltAz, SkyCoord)

from km3pipe.constants import orca_longitude, orca_latitude, orca_height

__author__ = "Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Moritz Lotze"
__email__ = "mlotze@km3net.de"
__status__ = "Development"


orca_loc = EarthLocation.from_geodetic(
    Longitude(orca_longitude * degree),
    Latitude(orca_latitude * degree),
    height=orca_height
)


def to_frame(time, zenith, azimuth, frame='galactic', unit=rad):
    """Tranform from the Detector frame to anything else.

    Parameters:
    -----------
    frame: str [default: 'galactic']
        The reference frame to transform to (needs to be known to astropy)

    """
    altitude = (0.5*np.pi - zenith) * unit
    azimuth *= unit
    orca_frame = AltAz(location=orca_loc, obstime=time)
    coords = SkyCoord(altitude, azimuth, frame=orca_frame)
    return coords.transform_to(frame)



#!/usr/bin/env python
# coding=utf-8
# Filename: jpp.py
# pylint: disable=
"""
Pump for the jpp file read through aanet interface.

"""
from __future__ import division, absolute_import, print_function

import numpy as np

from km3pipe.core import Pump, Blob
from km3pipe.dataclasses import (EventInfo, TimesliceFrameInfo,
                                 SummaryframeInfo, HitSeries,
                                 TimesliceHitSeries)
from km3pipe.logger import logging

log = logging.getLogger(__name__)  # pylint: disable=C0103

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = ["Thomas Heid"]
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class JPPPump(Pump):
    """A pump for JPP ROOT files."""

    def __init__(self, **context):
        super(self.__class__, self).__init__(**context)

        try:
            import jppy  # noqa
        except ImportError:
            raise ImportError("\nPlease install the jppy package:\n\n"
                              "    pip install jppy\n")

        self.event_index = self.get('index') or 0
        self.with_summaryslices = self.get('with_summaryslices') or False
        self.with_timeslice_hits = self.get('with_timeslice_hits') or False
        self.timeslice_index = 0
        self.timeslice_frame_index = 0
        self.summaryslice_index = 0
        self.summaryslice_frame_index = 0
        self.filename = self.get('filename')

        self.event_reader = jppy.PyJDAQEventReader(self.filename)
        self.timeslice_reader = jppy.PyJDAQTimesliceReader(self.filename)
        self.summaryslice_reader = jppy.PyJDAQSummarysliceReader(self.filename)
        self.blobs = self.blob_generator()

    def blob_generator(self):
        while self.with_timeslice_hits and self.timeslice_reader.has_next:
            self.timeslice_frame_index = 0
            self.timeslice_reader.retrieve_next_timeslice()
            while self.timeslice_reader.has_next_superframe:
                try:
                    yield self.extract_timeslice_frame()
                except IndexError:
                    log.warning("Skipping broken frame.")
                else:
                    self.timeslice_frame_index += 1
                finally:
                    self.timeslice_reader.retrieve_next_superframe()
            self.timeslice_index += 1

        while self.with_summaryslices and self.summaryslice_reader.has_next:
            self.summaryslice_frame_index = 0
            self.summaryslice_reader.retrieve_next_summaryslice()
            while self.summaryslice_reader.has_next_frame:
                yield self.extract_summaryslice_frame()
                self.summaryslice_reader.retrieve_next_frame()
                self.summaryslice_frame_index += 1
            self.summaryslice_index += 1

        while self.event_reader.has_next:
            yield self.extract_event()

        raise StopIteration

    def extract_event(self):
        blob = Blob()
        r = self.event_reader
        r.retrieve_next_event()  # do it at the beginning!

        n = r.number_of_snapshot_hits
        channel_ids = np.zeros(n, dtype='i')
        dom_ids = np.zeros(n, dtype='i')
        times = np.zeros(n, dtype='i')
        tots = np.zeros(n, dtype='i')
        triggereds = np.zeros(n, dtype='i')

        r.get_hits(channel_ids, dom_ids, times, tots, triggereds)

        nans = np.full(n, np.nan, dtype='<f8')
        hit_series = HitSeries.from_arrays(
            channel_ids, nans, nans, nans, dom_ids, np.arange(n), np.zeros(n),
            nans, nans, nans, nans, times, tots, triggereds, self.event_index
        )

        event_info = EventInfo((
            r.det_id, r.frame_index,
            0, # livetime_sec
            0, 0,  # MC ID and time
            0, # n_events_gen
            0, # n_files_gen
            r.overlays,
            # r.run_id,
            r.trigger_counter, r.trigger_mask,
            r.utc_nanoseconds, r.utc_seconds,
            np.nan, np.nan, np.nan,   # w1-w3
            self.event_index,
            ))

        self.event_index += 1
        blob['EventInfo'] = event_info
        blob['Hits'] = hit_series
        return blob

    def extract_timeslice_frame(self):
        blob = Blob()
        r = self.timeslice_reader
        n = r.number_of_hits
        channel_ids = np.zeros(n, dtype='i')
        dom_ids = np.zeros(n, dtype='i')
        times = np.zeros(n, dtype='i')
        tots = np.zeros(n, dtype='i')
        r.get_hits(channel_ids, dom_ids, times, tots)
        hit_series = TimesliceHitSeries.from_arrays(
            channel_ids, dom_ids, times, tots,
            self.timeslice_index, self.timeslice_frame_index
        )
        timesliceframe_info = TimesliceFrameInfo(
                r.dom_id,
                r.fifo_status,
                self.timeslice_frame_index,
                r.frame_index,
                r.has_udp_trailer,
                r.high_rate_veto,
                r.max_sequence_number,
                r.number_of_received_packets,
                self.timeslice_index,
                r.utc_nanoseconds,
                r.utc_seconds,
                r.white_rabbit_status,
                )

        blob['TimesliceHits'] = hit_series
        blob['TimesliceFrameInfo'] = timesliceframe_info
        return blob

    def extract_summaryslice_frame(self):
        blob = Blob()
        r = self.summaryslice_reader
        summaryframe_info = SummaryframeInfo(
                r.dom_id,
                r.fifo_status,
                self.summaryslice_frame_index,
                r.frame_index,
                r.has_udp_trailer,
                r.high_rate_veto,
                r.max_sequence_number,
                r.number_of_received_packets,
                self.summaryslice_index,
                r.utc_nanoseconds,
                r.utc_seconds,
                r.white_rabbit_status,
                )

        blob['SummaryframeInfo'] = summaryframe_info
        return blob

    def process(self, blob):
        return next(self.blobs)

    def __iter__(self):
        return self

    def next(self):
        """Python 2/3 compatibility for iterators"""
        return self.__next__()

    def __next__(self):
        return next(self.blobs)



# coding=utf-8
# Filename: mocks.py
"""
Mocks, fakes and dummies of external libraries.

"""
from __future__ import division, absolute_import, print_function


__author__ = "Tamas Gal and Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class FakeVec(object):
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

class FakeAanetHit(object):
    def __init__(self, channel_id, dir_x, dir_y, dir_z, dom_id, id, pmt_id,
                 pos_x, pos_y, pos_z, t0, t, tot, trig):
        # self.channel_id = chr(channel_id)
        self.channel_id = channel_id
        self.dir = FakeVec(dir_x, dir_y, dir_z)
        self.dom_id = dom_id
        self.id = id
        self.pmt_id = pmt_id
        self.pos = FakeVec(pos_x, pos_y, pos_z)
        self.t0 = t0
        self.t = t
        self.tot = tot
        self.trig = trig



# coding=utf-8
# Filename: k40calib.py
"""
Calculate t0set and write it to a CSV file.

Usage:
    k40calib FILE [-o OUTFILE]
    k40calib (-h | --help)
    k40calib --version

Options:
    -o OUTFILE   CSV file containing the t0 values.
    -h --help    Show this screen.
"""

from __future__ import division, absolute_import, print_function

from km3pipe import version

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Jonas Reubelt and Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


def k40calib(input_file, output_file=None):
    """K40 Calibration"""
    from km3modules import k40
    import km3pipe as kp
    import ROOT

    f = ROOT.TFile(input_file)
    dom_ids = [n.GetName().split('.')[0]
               for n in f.GetListOfKeys()
               if '.2S' in n.GetName()]
    detector = kp.hardware.Detector(det_id=14)
    if output_file is None:
        csv_filename = input_file + "_t0s.csv"
    else:
        csv_filename = output_file
    csv_file = open(csv_filename, 'w')
    csv_file.write("dom_id,tdc_channel,t0\n")
    for dom_id in dom_ids:
        print("Calibrating {0}...".format(dom_id))
        try:
            calibration = k40.calibrate_dom(dom_id, input_file, detector)
        except ValueError:
            print("   no data found, skipping.")
            continue
        t0set = calibration['opt_t0s'].x
        for tdc_channel, t0 in enumerate(t0set):
            csv_file.write("{0}, {1}, {2}\n"
                           .format(dom_id, tdc_channel, t0))
        print("    done.")
    csv_file.close()
    print("Calibration done, the t0 values were written to '{0}'."
          .format(csv_filename))


def main():
    from docopt import docopt
    args = docopt(__doc__, version=version)

    infile = args['FILE']
    k40calib(infile, args['-o'])
